#***************************************************************************
#*                                                                         *
#*   Copyright (c) 2021 Yorik van Havre <yorik@uncreated.net>              *
#*                                                                         *
#*   This program is free software; you can redistribute it and/or modify  *
#*   it under the terms of the GNU Lesser General Public License (LGPL)    *
#*   as published by the Free Software Foundation; either version 2 of     *
#*   the License, or (at your option) any later version.                   *
#*   for detail see the LICENCE text file.                                 *
#*                                                                         *
#*   This program is distributed in the hope that it will be useful,       *
#*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
#*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
#*   GNU Library General Public License for more details.                  *
#*                                                                         *
#*   You should have received a copy of the GNU Library General Public     *
#*   License along with this program; if not, write to the Free Software   *
#*   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
#*   USA                                                                   *
#*                                                                         *
#***************************************************************************


"""This module contains FreeCAD commands for the Draft workbench"""

import os
import FreeCAD as App
from draftutils.translate import translate, QT_TRANSLATE_NOOP
from draftgeoutils.general import geomType

from draftobjects.base import DraftObject


class Hatch(DraftObject):


    def __init__(self,obj):

        obj.Proxy = self
        self.setProperties(obj)

    def setProperties(self,obj):

        pl = obj.PropertiesList
        if not "Base" in pl:
            obj.addProperty("App::PropertyLink","Base","Hatch",
                            QT_TRANSLATE_NOOP("App::Property","The base object used by this object"))
        if not "File" in pl:
            obj.addProperty("App::PropertyFile","File","Hatch",
                            QT_TRANSLATE_NOOP("App::Property","The PAT file used by this object"))
        if not "Pattern" in pl:
            obj.addProperty("App::PropertyString","Pattern","Hatch",
                            QT_TRANSLATE_NOOP("App::Property","The pattern name used by this object"))
        if not "Scale" in pl:
            obj.addProperty("App::PropertyFloat","Scale","Hatch",
                            QT_TRANSLATE_NOOP("App::Property","The pattern scale used by this object"))
        if not "Rotation" in pl:
            obj.addProperty("App::PropertyAngle","Rotation","Hatch",
                            QT_TRANSLATE_NOOP("App::Property","The pattern rotation used by this object"))
        if not "Translate" in pl:
            obj.addProperty("App::PropertyBool","Translate","Hatch",
                            QT_TRANSLATE_NOOP("App::Property","If set to False, hatch is applied as is to the faces, without translation (this might give wrong results for non-XY faces)"))
            obj.Translate = True
        self.Type = "Hatch"

    def onDocumentRestored(self,obj):

        self.setProperties(obj)

    def dumps(self):

        return None

    def loads(self,state):

        return None

    def execute(self,obj):

        if self.props_changed_placement_only(obj) \
                or not obj.Base \
                or not obj.File \
                or not obj.Pattern \
                or not obj.Scale \
                or not obj.Pattern in self.getPatterns(obj.File) \
                or not obj.Base.isDerivedFrom("Part::Feature") \
                or not obj.Base.Shape.Faces:
            self.props_changed_clear()
            return

        import Part
        import TechDraw

        shapes = []
        for face in obj.Base.Shape.Faces:
            if face.findPlane(): # Only planar faces.
                face = face.copy()
                if obj.Translate:
                    mtx = None
                    w = face.normalAt(0, 0)
                    # Try to base a matrix on the first straight edge with
                    # a reasonable length (> 0.001):
                    for e in face.Edges:
                        if geomType(e) == "Line":
                            sta = e.firstVertex().Point
                            end = e.lastVertex().Point
                            u = end.sub(sta)
                            if u.Length > 0.001:
                                u = u.normalize()
                                v = w.cross(u)
                                mtx = App.Matrix(u.x, v.x, w.x, sta.x,
                                                 u.y, v.y, w.y, sta.y,
                                                 u.z, v.z, w.z, sta.z,
                                                 0.0, 0.0, 0.0, 1.0)
                                break
                    # If no suitable straight edge was found use a default matrix:
                    if not mtx:
                        cen = face.CenterOfMass
                        rot = App.Rotation(App.Vector(0,0,1), w)
                        mtx = App.Placement(cen, rot).Matrix
                    face = face.transformGeometry(mtx.inverse()).Faces[0]
                if obj.Rotation.Value:
                    face.rotate(App.Vector(), App.Vector(0,0,1), -obj.Rotation)
                shape = TechDraw.makeGeomHatch(face, obj.Scale, obj.Pattern, obj.File)
                if obj.Rotation.Value:
                    shape.rotate(App.Vector(), App.Vector(0,0,1), obj.Rotation)
                if obj.Translate:
                    shape = shape.transformGeometry(mtx)
                shapes.append(shape)
        if shapes:
            obj.Shape = Part.makeCompound(shapes)
        self.props_changed_clear()

    def onChanged(self, obj, prop):

        self.props_changed_store(prop)

    def getPatterns(self,filename):

        """returns a list of pattern names found in a PAT file"""
        patterns = []
        if os.path.exists(filename):
            with open(filename) as patfile:
                for line in patfile:
                    if line.startswith("*"):
                        patterns.append(line.split(",")[0][1:])
        return patterns



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2022-2023 FreeCAD Project Association                   *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

from PySide import QtCore, QtWidgets

from AddonManagerTest.app.mocks import SignalCatcher


class DialogInteractor(QtCore.QObject):
    """Takes the title of the dialog and a callable. The callable is passed the widget
    we found and can do whatever it wants to it. Whatever it does should eventually
    close the dialog, however."""

    def __init__(self, dialog_to_watch_for, interaction):
        super().__init__()

        # Status variables for tests to check:
        self.dialog_found = False
        self.has_run = False
        self.button_found = False
        self.interaction = interaction

        self.dialog_to_watch_for = dialog_to_watch_for

        self.execution_counter = 0
        self.timer = QtCore.QTimer()
        self.timer.timeout.connect(self.run)
        self.timer.start(10)

    def run(self):
        widget = QtWidgets.QApplication.activeModalWidget()
        if widget and self._dialog_matches(widget):
            # Found the dialog we are looking for: now try to run the interaction
            if self.interaction is not None and callable(self.interaction):
                self.interaction(widget)
            self.dialog_found = True
            self.timer.stop()

        self.has_run = True
        self.execution_counter += 1
        if self.execution_counter > 100:
            print("Stopped timer after 100 iterations")
            self.timer.stop()

    def _dialog_matches(self, widget) -> bool:
        # Is this the widget we are looking for? Only applies on Linux and Windows: macOS
        # doesn't set the title of a modal dialog:
        os = QtCore.QSysInfo.productType()  # Qt5 gives "osx", Qt6 gives "macos"
        if os in ["osx", "macos"] or (
            hasattr(widget, "windowTitle")
            and callable(widget.windowTitle)
            and widget.windowTitle() == self.dialog_to_watch_for
        ):
            return True
        return False


class DialogWatcher(DialogInteractor):
    """Examine the running GUI and look for a modal dialog with a given title, containing a button
    with a role. Click that button, which is expected to close the dialog. Generally run on
    a one-shot QTimer to allow the dialog time to open up. If the specified dialog is found, but
    it does not contain the expected button, button_found will be false, and the dialog will be
    closed with a reject() slot."""

    def __init__(self, dialog_to_watch_for, button=QtWidgets.QDialogButtonBox.NoButton):
        super().__init__(dialog_to_watch_for, self.click_button)
        if button != QtWidgets.QDialogButtonBox.NoButton:
            self.button = button
        else:
            self.button = QtWidgets.QDialogButtonBox.Cancel

    def click_button(self, widget):
        button_boxes = widget.findChildren(QtWidgets.QDialogButtonBox)
        if len(button_boxes) == 1:  # There should be one, and only one
            button_to_click = button_boxes[0].button(self.button)
            if button_to_click:
                self.button_found = True
                button_to_click.click()
            else:
                widget.reject()
        else:
            widget.reject()


class FakeWorker:
    def __init__(self):
        self.called = False
        self.should_continue = True

    def work(self):
        while self.should_continue:
            QtCore.QCoreApplication.processEvents(QtCore.QEventLoop.AllEvents, 100)

    def stop(self):
        self.should_continue = False


class MockThread:
    def wait(self):
        pass

    def isRunning(self):
        return False


class AsynchronousMonitor:
    """Watch for a signal to be emitted for at most some given number of milliseconds"""

    def __init__(self, signal):
        self.signal = signal
        self.signal_catcher = SignalCatcher()
        self.signal.connect(self.signal_catcher.catch_signal)
        self.kill_timer = QtCore.QTimer()
        self.kill_timer.setSingleShot(True)
        self.kill_timer.timeout.connect(self.signal_catcher.die)

    def wait_for_at_most(self, max_wait_millis) -> None:
        self.kill_timer.setInterval(max_wait_millis)
        self.kill_timer.start()
        while not self.signal_catcher.caught and not self.signal_catcher.killed:
            QtCore.QCoreApplication.processEvents(QtCore.QEventLoop.AllEvents, 10)
        self.kill_timer.stop()

    def good(self) -> bool:
        return self.signal_catcher.caught and not self.signal_catcher.killed



# ***************************************************************************
# *   Copyright (c) 2014 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Lesser General Public License for more details.                   *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
from FreeCAD import Units
import Path
import argparse
import datetime
import shlex
import Path.Post.Utils as PostUtils
import PathScripts.PathUtils as PathUtils

TOOLTIP = """
This is a postprocessor file for the Path workbench. It is used to
take a pseudo-G-code fragment outputted by a Path object, and output
real G-code suitable for a linuxcnc 3 axis mill. This postprocessor, once placed
in the appropriate PathScripts folder, can be used directly from inside
FreeCAD, via the GUI importer or via python scripts with:

import linuxcnc_post
linuxcnc_post.export(object,"/path/to/file.ncc","")
"""

now = datetime.datetime.now()

parser = argparse.ArgumentParser(prog="linuxcnc", add_help=False)
parser.add_argument("--no-header", action="store_true", help="suppress header output")
parser.add_argument(
    "--no-comments", action="store_true", help="suppress comment output"
)
parser.add_argument(
    "--line-numbers", action="store_true", help="prefix with line numbers"
)
parser.add_argument(
    "--no-show-editor",
    action="store_true",
    help="don't pop up editor before writing output",
)
parser.add_argument(
    "--precision", default="3", help="number of digits of precision, default=3"
)
parser.add_argument(
    "--preamble",
    help='set commands to be issued before the first command, default="G17\nG90"',
)
parser.add_argument(
    "--postamble",
    help='set commands to be issued after the last command, default="M05\nG17 G90\nM2"',
)
parser.add_argument(
    "--inches", action="store_true", help="Convert output for US imperial mode (G20)"
)
parser.add_argument(
    "--modal",
    action="store_true",
    help="Output the Same G-command Name USE NonModal Mode",
)
parser.add_argument(
    "--axis-modal", action="store_true", help="Output the Same Axis Value Mode"
)
parser.add_argument(
    "--no-tlo",
    action="store_true",
    help="suppress tool length offset (G43) following tool changes",
)

TOOLTIP_ARGS = parser.format_help()

# These globals set common customization preferences
OUTPUT_COMMENTS = True
OUTPUT_HEADER = True
OUTPUT_LINE_NUMBERS = False
SHOW_EDITOR = True
MODAL = False  # if true commands are suppressed if the same as previous line.
USE_TLO = True  # if true G43 will be output following tool changes
OUTPUT_DOUBLES = (
    True  # if false duplicate axis values are suppressed if the same as previous line.
)
COMMAND_SPACE = " "
LINENR = 100  # line number starting value

# These globals will be reflected in the Machine configuration of the project
UNITS = "G21"  # G21 for metric, G20 for us standard
UNIT_SPEED_FORMAT = "mm/min"
UNIT_FORMAT = "mm"

MACHINE_NAME = "LinuxCNC"
CORNER_MIN = {"x": 0, "y": 0, "z": 0}
CORNER_MAX = {"x": 500, "y": 300, "z": 300}
PRECISION = 3

# Preamble text will appear at the beginning of the GCODE output file.
PREAMBLE = """G17 G54 G40 G49 G80 G90
"""

# Postamble text will appear following the last operation.
POSTAMBLE = """M05
G17 G54 G90 G80 G40
M2
"""

# Pre operation text will be inserted before every operation
PRE_OPERATION = """"""

# Post operation text will be inserted after every operation
POST_OPERATION = """"""

# Tool Change commands will be inserted before a tool change
TOOL_CHANGE = """"""

# to distinguish python built-in open function from the one declared below
if open.__module__ in ["__builtin__", "io"]:
    pythonopen = open


def processArguments(argstring):
    global OUTPUT_HEADER
    global OUTPUT_COMMENTS
    global OUTPUT_LINE_NUMBERS
    global SHOW_EDITOR
    global PRECISION
    global PREAMBLE
    global POSTAMBLE
    global UNITS
    global UNIT_SPEED_FORMAT
    global UNIT_FORMAT
    global MODAL
    global USE_TLO
    global OUTPUT_DOUBLES

    try:
        args = parser.parse_args(shlex.split(argstring))
        if args.no_header:
            OUTPUT_HEADER = False
        if args.no_comments:
            OUTPUT_COMMENTS = False
        if args.line_numbers:
            OUTPUT_LINE_NUMBERS = True
        if args.no_show_editor:
            SHOW_EDITOR = False
        print("Show editor = %d" % SHOW_EDITOR)
        PRECISION = args.precision
        if args.preamble is not None:
            PREAMBLE = args.preamble
        if args.postamble is not None:
            POSTAMBLE = args.postamble
        if args.inches:
            UNITS = "G20"
            UNIT_SPEED_FORMAT = "in/min"
            UNIT_FORMAT = "in"
            PRECISION = 4
        if args.modal:
            MODAL = True
        if args.no_tlo:
            USE_TLO = False
        if args.axis_modal:
            print("here")
            OUTPUT_DOUBLES = False

    except Exception:
        return False

    return True


def export(objectslist, filename, argstring):
    if not processArguments(argstring):
        return None
    global UNITS
    global UNIT_FORMAT
    global UNIT_SPEED_FORMAT

    for obj in objectslist:
        if not hasattr(obj, "Path"):
            print(
                "the object "
                + obj.Name
                + " is not a path. Please select only path and Compounds."
            )
            return None

    print("postprocessing...")
    gcode = ""

    # write header
    if OUTPUT_HEADER:
        gcode += linenumber() + "(Exported by FreeCAD)\n"
        gcode += linenumber() + "(Post Processor: " + __name__ + ")\n"
        gcode += linenumber() + "(Output Time:" + str(now) + ")\n"

    # Write the preamble
    if OUTPUT_COMMENTS:
        gcode += linenumber() + "(begin preamble)\n"
    for line in PREAMBLE.splitlines(False):
        gcode += linenumber() + line + "\n"
    gcode += linenumber() + UNITS + "\n"

    for obj in objectslist:

        # Skip inactive operations
        if hasattr(obj, "Active"):
            if not obj.Active:
                continue
        if hasattr(obj, "Base") and hasattr(obj.Base, "Active"):
            if not obj.Base.Active:
                continue

        # do the pre_op
        if OUTPUT_COMMENTS:
            gcode += linenumber() + "(begin operation: %s)\n" % obj.Label
            gcode += linenumber() + "(machine units: %s)\n" % (UNIT_SPEED_FORMAT)
        for line in PRE_OPERATION.splitlines(True):
            gcode += linenumber() + line

        # get coolant mode
        coolantMode = "None"
        if (
            hasattr(obj, "CoolantMode")
            or hasattr(obj, "Base")
            and hasattr(obj.Base, "CoolantMode")
        ):
            if hasattr(obj, "CoolantMode"):
                coolantMode = obj.CoolantMode
            else:
                coolantMode = obj.Base.CoolantMode

        # turn coolant on if required
        if OUTPUT_COMMENTS:
            if not coolantMode == "None":
                gcode += linenumber() + "(Coolant On:" + coolantMode + ")\n"
        if coolantMode == "Flood":
            gcode += linenumber() + "M8" + "\n"
        if coolantMode == "Mist":
            gcode += linenumber() + "M7" + "\n"

        # process the operation gcode
        gcode += parse(obj)

        # do the post_op
        if OUTPUT_COMMENTS:
            gcode += linenumber() + "(finish operation: %s)\n" % obj.Label
        for line in POST_OPERATION.splitlines(True):
            gcode += linenumber() + line

        # turn coolant off if required
        if not coolantMode == "None":
            if OUTPUT_COMMENTS:
                gcode += linenumber() + "(Coolant Off:" + coolantMode + ")\n"
            gcode += linenumber() + "M9" + "\n"

    # do the post_amble
    if OUTPUT_COMMENTS:
        gcode += "(begin postamble)\n"
    for line in POSTAMBLE.splitlines(True):
        gcode += linenumber() + line

    if FreeCAD.GuiUp and SHOW_EDITOR:
        final = gcode
        if len(gcode) > 100000:
            print("Skipping editor since output is greater than 100kb")
        else:
            dia = PostUtils.GCodeEditorDialog()
            dia.editor.setText(gcode)
            result = dia.exec_()
            if result:
                final = dia.editor.toPlainText()
    else:
        final = gcode

    print("done postprocessing.")

    if not filename == "-":
        gfile = pythonopen(filename, "w")
        gfile.write(final)
        gfile.close()

    return final


def linenumber():
    global LINENR
    if OUTPUT_LINE_NUMBERS is True:
        LINENR += 10
        return "N" + str(LINENR) + " "
    return ""


def parse(pathobj):
    global PRECISION
    global MODAL
    global OUTPUT_DOUBLES
    global UNIT_FORMAT
    global UNIT_SPEED_FORMAT

    out = ""
    lastcommand = None
    precision_string = "." + str(PRECISION) + "f"
    currLocation = {}  # keep track for no doubles

    # the order of parameters
    # linuxcnc doesn't want K properties on XY plane  Arcs need work.
    params = [
        "X",
        "Y",
        "Z",
        "A",
        "B",
        "C",
        "I",
        "J",
        "F",
        "S",
        "T",
        "Q",
        "R",
        "L",
        "H",
        "D",
        "P",
    ]
    firstmove = Path.Command("G0", {"X": -1, "Y": -1, "Z": -1, "F": 0.0})
    currLocation.update(firstmove.Parameters)  # set First location Parameters

    if hasattr(pathobj, "Group"):  # We have a compound or project.
        # if OUTPUT_COMMENTS:
        #     out += linenumber() + "(compound: " + pathobj.Label + ")\n"
        for p in pathobj.Group:
            out += parse(p)
        return out
    else:  # parsing simple path

        # groups might contain non-path things like stock.
        if not hasattr(pathobj, "Path"):
            return out

        # if OUTPUT_COMMENTS:
        #     out += linenumber() + "(" + pathobj.Label + ")\n"

        for c in PathUtils.getPathWithPlacement(pathobj).Commands:

            outstring = []
            command = c.Name
            outstring.append(command)

            # if modal: suppress the command if it is the same as the last one
            if MODAL is True:
                if command == lastcommand:
                    outstring.pop(0)

            if c.Name[0] == "(" and not OUTPUT_COMMENTS:  # command is a comment
                continue

            # Now add the remaining parameters in order
            for param in params:
                if param in c.Parameters:
                    if param == "F" and (
                        currLocation[param] != c.Parameters[param] or OUTPUT_DOUBLES
                    ):
                        if c.Name not in [
                            "G0",
                            "G00",
                        ]:  # linuxcnc doesn't use rapid speeds
                            speed = Units.Quantity(
                                c.Parameters["F"], FreeCAD.Units.Velocity
                            )
                            if speed.getValueAs(UNIT_SPEED_FORMAT) > 0.0:
                                outstring.append(
                                    param
                                    + format(
                                        float(speed.getValueAs(UNIT_SPEED_FORMAT)),
                                        precision_string,
                                    )
                                )
                        else:
                            continue
                    elif param == "T":
                        outstring.append(param + str(int(c.Parameters["T"])))
                    elif param == "H":
                        outstring.append(param + str(int(c.Parameters["H"])))
                    elif param == "D":
                        outstring.append(param + str(int(c.Parameters["D"])))
                    elif param == "S":
                        outstring.append(param + str(int(c.Parameters["S"])))
                    else:
                        if (
                            (not OUTPUT_DOUBLES)
                            and (param in currLocation)
                            and (currLocation[param] == c.Parameters[param])
                        ):
                            continue
                        else:
                            pos = Units.Quantity(
                                c.Parameters[param], FreeCAD.Units.Length
                            )
                            outstring.append(
                                param
                                + format(
                                    float(pos.getValueAs(UNIT_FORMAT)), precision_string
                                )
                            )

            # store the latest command
            lastcommand = command
            currLocation.update(c.Parameters)

            # Check for Tool Change:
            if command == "M6":
                # stop the spindle
                out += linenumber() + "M5\n"
                for line in TOOL_CHANGE.splitlines(True):
                    out += linenumber() + line

                # add height offset
                if USE_TLO:
                    tool_height = "\nG43 H" + str(int(c.Parameters["T"]))
                    outstring.append(tool_height)

            if command == "message":
                if OUTPUT_COMMENTS is False:
                    out = []
                else:
                    outstring.pop(0)  # remove the command

            # prepend a line number and append a newline
            if len(outstring) >= 1:
                if OUTPUT_LINE_NUMBERS:
                    outstring.insert(0, (linenumber()))

                # append the line to the final output
                for w in outstring:
                    out += w + COMMAND_SPACE
                # Note: Do *not* strip `out`, since that forces the allocation
                # of a contiguous string & thus quadratic complexity.
                out += "\n"

        return out


# print(__name__ + " gcode postprocessor loaded.")



# ***************************************************************************
# *   Copyright (c) 2014 sliptonic <shopinthewoods@gmail.com>               *
# *                 2023 Nicklas SB Karlsson <nk@nksb.eu>                   *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Lesser General Public License for more details.                   *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************/
# *                                                                         *
# *  Modified 2021 by lcorley to support Crossfire(tm) plasma tables        *
# *  Made a copy on 2023-01-17 of crossfire_post.py with name               *
# *       fangling_post.py                                                  *
# *  Modified 2023-01-17- by Nicklas SB Karlsson to support FangLing plasma *
# *           cutters. Still untested on real cutter.                       *
# *                                                                         *
# ***************************************************************************/
from __future__ import print_function
import FreeCAD
from FreeCAD import Units
import Path
import argparse
import datetime
import shlex
#from PathScripts import PostUtils
import Path.Post.Utils as PostUtils
from PathScripts import PathUtils

TOOLTIP = '''
This is a postprocessor file for the Path workbench. It is used to
take a pseudo-G-code fragment outputted by a Path object, and output
real G-code suitable for FangLing plasma cutters.
This postprocessor, once placed in the appropriate PathScripts folder,
can be used directly from inside FreeCAD, via the GUI importer or
via python scripts with:

import FangLing_post
FangLing_post.export(object,"/path/to/file.ncc","")
'''

now = datetime.datetime.now()

parser = argparse.ArgumentParser(prog='FangLing', add_help=False)
parser.add_argument('--no-header', action='store_true', help='suppress header output')
parser.add_argument('--no-comments', action='store_true', help='suppress comment output')
parser.add_argument('--line-numbers', action='store_true', help='prefix with line numbers')
parser.add_argument('--no-show-editor', action='store_true', help='don\'t pop up editor before writing output')
parser.add_argument('--precision', default='3', help='number of digits of precision, default=3')
parser.add_argument('--preamble', help='set commands to be issued before the first command, default="G17\nG90"')
parser.add_argument('--postamble', help='set commands to be issued after the last command, default="M5\nG17 G90\nM2"')
parser.add_argument('--inches', action='store_true', help='Convert output for US imperial mode (G20)')
parser.add_argument('--modal', action='store_true', help='Output the Same G-command Name USE NonModal Mode')
parser.add_argument('--axis-modal', action='store_true', help='Output the Same Axis Value Mode')
parser.add_argument('--no-tlo', action='store_true', help='suppress tool length offset (G43) following tool changes')

TOOLTIP_ARGS = parser.format_help()

# These globals set common customization preferences
OUTPUT_COMMENTS = True
OUTPUT_HEADER = True
OUTPUT_LINE_NUMBERS = False
SHOW_EDITOR = True
MODAL = False  # if true commands are suppressed if the same as previous line.
USE_TLO = False # if true G43 will be output following tool changes
OUTPUT_DOUBLES = True  # if false duplicate axis values are suppressed if the same as previous line.
COMMAND_SPACE = " "
LINENR = 100  # line number starting value

# These globals will be reflected in the Machine configuration of the project
UNITS = "G21"  # G21 for metric, G20 for us standard
UNIT_SPEED_FORMAT = 'mm/min'
UNIT_FORMAT = 'mm'

MACHINE_NAME = "FangLing"
CORNER_MIN = {'x': 0, 'y': 0, 'z': 0}
CORNER_MAX = {'x': 500, 'y': 300, 'z': 300}
PRECISION = 3

# Preamble text will appear at the beginning of the GCODE output file.
#PREAMBLE = '''G17 G54 G40 G49 G80 G90'''
#Changed to match preamble produced by Fusion 360
PREAMBLE = '''G90
'''

# Postamble text will appear following the last operation.
POSTAMBLE = '''M8
G90 G40
M2
'''

# Pre operation text will be inserted before every operation
PRE_OPERATION = ''''''

# Post operation text will be inserted after every operation
POST_OPERATION = ''''''

# Tool Change commands will be inserted before a tool change
TOOL_CHANGE = ''''''

# to distinguish python built-in open function from the one declared below
if open.__module__ in ['__builtin__','io']:
    pythonopen = open


def processArguments(argstring):
    # pylint: disable=global-statement
    global OUTPUT_HEADER
    global OUTPUT_COMMENTS
    global OUTPUT_LINE_NUMBERS
    global SHOW_EDITOR
    global PRECISION
    global PREAMBLE
    global POSTAMBLE
    global UNITS
    global UNIT_SPEED_FORMAT
    global UNIT_FORMAT
    global MODAL
    global USE_TLO
    global OUTPUT_DOUBLES

    try:
        args = parser.parse_args(shlex.split(argstring))
        if args.no_header:
            OUTPUT_HEADER = False
        if args.no_comments:
            OUTPUT_COMMENTS = False
        if args.line_numbers:
            OUTPUT_LINE_NUMBERS = True
        if args.no_show_editor:
            SHOW_EDITOR = False
        print("Show editor = %d" % SHOW_EDITOR)
        PRECISION = args.precision
        if args.preamble is not None:
            PREAMBLE = args.preamble
        if args.postamble is not None:
            POSTAMBLE = args.postamble
        if args.inches:
            UNITS = 'G20'
            UNIT_SPEED_FORMAT = 'in/min'
            UNIT_FORMAT = 'in'
            PRECISION = 4
        if args.modal:
            MODAL = True
        if args.no_tlo:
            USE_TLO = False
        if args.axis_modal:
            print ('here')
            OUTPUT_DOUBLES = False

    except Exception: # pylint: disable=broad-except
        return False

    return True


def export(objectslist, filename, argstring):
    # pylint: disable=global-statement
    if not processArguments(argstring):
        return None
    global UNITS
    global UNIT_FORMAT
    global UNIT_SPEED_FORMAT
    global HORIZRAPID
    global VERTRAPID

    for obj in objectslist:
        if not hasattr(obj, "Path"):
            print("the object " + obj.Name + " is not a path. Please select only path and Compounds.")
            return None

    print("postprocessing...")
    gcode = ""

    # write header
    if OUTPUT_HEADER:
        gcode += linenumber() + "(Exported by FreeCAD)\n"
        gcode += linenumber() + "(Post Processor: " + __name__ + ")\n"
        gcode += linenumber() + "(Output Time:" + str(now) + ")\n"

    # Write the preamble
    if OUTPUT_COMMENTS:
        gcode += linenumber() + "(begin preamble)\n"
    for line in PREAMBLE.splitlines(False):
        gcode += linenumber() + line + "\n"
    gcode += linenumber() + UNITS + "\n"

    gcode += linenumber() + "G92 X0 Y0\n"

    for obj in objectslist:

        # Skip inactive operations
        if hasattr(obj, 'Active'):
            if not obj.Active:
                continue
        if hasattr(obj, 'Base') and hasattr(obj.Base, 'Active'):
            if not obj.Base.Active:
                continue

        # fetch machine details
        job = PathUtils.findParentJob(obj)

        if hasattr(job, "MachineUnits"):
            if job.MachineUnits == "Metric":
                UNITS = "G21"
                UNIT_FORMAT = 'mm'
                UNIT_SPEED_FORMAT = 'mm/min'
            elif job.MachineUnits == "Imperial":
                UNITS = "G20"
                UNIT_FORMAT = 'in'
                UNIT_SPEED_FORMAT = 'in/min'
            else:
                print("Machine units " + job.MachineUnits + " could not be handled.")
                return None

        if hasattr(job, "SetupSheet"):
            if hasattr(job.SetupSheet, "HorizRapid"):
                HORIZRAPID = Units.Quantity(job.SetupSheet.HorizRapid, FreeCAD.Units.Velocity)
            if hasattr(job.SetupSheet, "VertRapid"):
                VERTRAPID = Units.Quantity(job.SetupSheet.HorizRapid, FreeCAD.Units.Velocity)

        # process the operation gcode
        gcode += parse(obj)

    # do the post_amble
    if OUTPUT_COMMENTS:
        gcode += "(begin postamble)\n"
    for line in POSTAMBLE.splitlines(True):
        gcode += linenumber() + line

    if FreeCAD.GuiUp and SHOW_EDITOR:
        dia = PostUtils.GCodeEditorDialog()
        dia.editor.setText(gcode)
        if dia.exec_():
            final = dia.editor.toPlainText()
        else:
            final = gcode
    else:
        final = gcode

    print("done postprocessing.")

    writeFile(filename, final)

    return final


def linenumber():
    # pylint: disable=global-statement
    global LINENR
    if OUTPUT_LINE_NUMBERS:
        LINENR += 10
        return "N" + str(LINENR) + " "
    return ""


def parse(pathobj):
    # pylint: disable=global-statement
    global PRECISION
    global MODAL
    global OUTPUT_DOUBLES
    global UNIT_FORMAT
    global UNIT_SPEED_FORMAT

    out = ""
    #print(pathobj.Path.toGCode())
    lastcommand = None
    precision_string = '.' + str(PRECISION) + 'f' # 'g' instead of 'f' removes trailing zeroes
    currLocation = {}  # keep track for no doubles

    # the order of parameters
    # mach3_4 doesn't want K properties on XY plane  Arcs need work.
    #params = ['X', 'Y', 'Z', 'A', 'B', 'C', 'I', 'J', 'F', 'S', 'T', 'Q', 'R', 'L', 'H', 'D', 'P']
    # FangLing plasma cutters doesn't use Z values
    params = ['X', 'Y', 'A', 'B', 'C', 'I', 'J', 'F', 'S', 'T', 'Q', 'R', 'L', 'H', 'D', 'P'] # It have not been checked which parameters FangLing accept
    firstmove = Path.Command("G0", {"X": -1, "Y": -1, "Z": -1, "F": 0.0})
    currLocation.update(firstmove.Parameters)  # set First location Parameters

    if hasattr(pathobj, "Group"):  # We have a compound or project.
        # if OUTPUT_COMMENTS:
        #     out += linenumber() + "(compound: " + pathobj.Label + ")\n"
        for p in pathobj.Group:
            out += parse(p)
        return out
    else:  # parsing simple path

        # groups might contain non-path things like stock.
        if not hasattr(pathobj, "Path"):
            return out

        # if OUTPUT_COMMENTS:
        #     out += linenumber() + "(" + pathobj.Label + ")\n"

        adaptiveOp = False
        opHorizRapid = 0
        opVertRapid = 0

        if 'Adaptive' in pathobj.Name:
            adaptiveOp = True
            if hasattr(pathobj, 'ToolController'):
                if hasattr(pathobj.ToolController, 'HorizRapid') and pathobj.ToolController.HorizRapid > 0:
                    opHorizRapid = Units.Quantity(pathobj.ToolController.HorizRapid, FreeCAD.Units.Velocity)
                else:
                    FreeCAD.Console.PrintWarning('Tool Controller Horizontal Rapid Values are unset'+ '\n')

                if hasattr(pathobj.ToolController, 'VertRapid') and pathobj.ToolController.VertRapid > 0:
                    opVertRapid = Units.Quantity(pathobj.ToolController.VertRapid, FreeCAD.Units.Velocity)
                else:
                    FreeCAD.Console.PrintWarning('Tool Controller Vertical Rapid Values are unset'+ '\n')

        #print(pathobj.Path.Commands)
        for c in pathobj.Path.Commands:

            outstring = []
            command = c.Name

            if c.Name in ["G0", "G00"] and lastcommand in ["G1", "G2", "G3"]:
                # A rapid move following a positioning move indicates the cut is complete
                # print ("end of cut")
                if OUTPUT_COMMENTS:
                    out += linenumber() + " (Torch Off)\n"
                out += linenumber() + " M08\n"
            if c.Name in ["G1", "G2", "G3"] and lastcommand in ["G0", "G00"]:
                # A positioning move following a rapid move indicates a new cut is starting
                # print ("new cut")
                if OUTPUT_COMMENTS:
                    out += linenumber() + " (Torch On)\n"
                out += linenumber() + " M07\n"
                if OUTPUT_COMMENTS:
                    out += linenumber() + " (Pierce Delay)\n"
                out += linenumber() + " G4 P1\n"

            if adaptiveOp and c.Name in ["G0", "G00"]:
                if opHorizRapid and opVertRapid:
                    command = 'G1'
                else:
                    outstring.append('(Tool Controller Rapid Values are unset)' + '\n')

            # Filter out vertical moves since FangLing plasma cutters does not support them
            if command in ["G0", "G00"] and len(c.Parameters) == 1 and 'Z' in c.Parameters:
                #print(c, c.Name, len(c.Parameters), list(c.Parameters))
                print('vertical move detected and removed')
            if command in ["G54"]:
                #print(c, c.Name, len(c.Parameters), list(c.Parameters))
                print('Set coordinate system removed')
            else:
                if command in ["G0", "G00"] and not 'X' in c.Parameters and not 'Y' in c.Parameters:
                    print('No horizontal move detected so removed')
                else:
                    outstring.append(command)

            # if modal: suppress the command if it is the same as the last one
            if MODAL is True:
                if command == lastcommand:
                    outstring.pop(0)

            if c.Name[0] == '(' and not OUTPUT_COMMENTS: # command is a comment
                continue

            # Now add the remaining parameters in order
            for param in params:
                if param in c.Parameters:
                    if param == 'F':
                        continue
                    elif param in ['T', 'H', 'D', 'S']:
                        outstring.append(param + str(int(c.Parameters[param])))
                    elif not OUTPUT_DOUBLES and param in currLocation and currLocation[param] == c.Parameters[param]:
                        continue
                    else:
                        pos = Units.Quantity(c.Parameters[param], FreeCAD.Units.Length)
                        outstring.append(
                            param + format(float(pos.getValueAs(UNIT_FORMAT)), precision_string))

            # store the latest command
            lastcommand = command
            currLocation.update(c.Parameters)

            # Check for Tool Change:
            # Tool changes not supported on FangLing plasma cutters
            if command == 'M6':
                outstring.pop(0) # remove the M6
                outstring.pop(0) # remove the tool number
                # stop the spindle
                #out += linenumber() + " M5\n"
                #for line in TOOL_CHANGE.splitlines(True):
                #out += linenumber() + line

                # add height offset
                # No TLO available in g-code
                #if USE_TLO:
                #    tool_height = '\nG43 H' + str(int(c.Parameters['T']))
                #    outstring.append(tool_height)
                print('M6 tool change removed')

            if command == "message":
                if not OUTPUT_COMMENTS:
                    out = []
                else:
                    outstring.pop(0)  # remove the command

            # prepend a line number and append a newline
            if len(outstring) >= 1:
                if OUTPUT_LINE_NUMBERS:
                    outstring.insert(0, (linenumber()))

                # append the line to the final output
                for w in outstring:
                    out += w + COMMAND_SPACE
                out = out.strip() + "\n"

        return out

# print(__name__ + " gcode postprocessor loaded.")


def writeFile(filename, final):
    if not filename == '-':
        gfile = pythonopen(filename, "w")
        gfile.write(final)
        gfile.close()

    return



# ***************************************************************************
# *   (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>                  *
# *   (c) 2009, 2010 Ken Cline <cline@frii.com>                             *
# *   (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de>           *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides GUI tools to create text shapes with a particular font.

These text shapes are made of various edges and closed faces, and therefore
can be extruded to create solid bodies that can be used in boolean
operations. That is, these text shapes can be used for engraving text
into solid bodies.

They are more complex that simple text annotations.
"""
## @package gui_shapestrings
# \ingroup draftguitools
# \brief Provides GUI tools to create text shapes with a particular font.

## \addtogroup draftguitools
# @{
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCAD as App
import FreeCADGui as Gui
import Draft_rc
import DraftVecUtils
import draftutils.utils as utils
import draftguitools.gui_base_original as gui_base_original
import draftguitools.gui_tool_utils as gui_tool_utils
import draftutils.todo as todo

from drafttaskpanels.task_shapestring import ShapeStringTaskPanelCmd
from draftutils.translate import translate
from draftutils.messages import _msg, _err

# The module is used to prevent complaints from code checkers (flake8)
True if Draft_rc.__name__ else False


class ShapeString(gui_base_original.Creator):
    """Gui command for the ShapeString tool."""

    def GetResources(self):
        """Set icon, menu and tooltip."""

        d = {'Pixmap': 'Draft_ShapeString',
             'MenuText': QT_TRANSLATE_NOOP("Draft_ShapeString", "Shape from text"),
             'ToolTip': QT_TRANSLATE_NOOP("Draft_ShapeString", "Creates a shape from a text string by choosing a specific font and a placement.\nThe closed shapes can be used for extrusions and boolean operations.")}
        return d

    def Activated(self):
        """Execute when the command is called."""
        super(ShapeString, self).Activated(name="ShapeString")
        if self.ui:
            self.ui.sourceCmd = self
            self.task = ShapeStringTaskPanelCmd(self)
            self.call = self.view.addEventCallback("SoEvent", self.task.action)
            _msg(translate("draft", "Pick ShapeString location point"))
            todo.ToDo.delay(Gui.Control.showDialog, self.task)


Gui.addCommand('Draft_ShapeString', ShapeString())

## @}



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2022 FreeCAD Project Association                        *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

""" Contains a wrapper class for a table listing authors and maintainers """

import os

from PySide.QtWidgets import QTableWidgetItem
from PySide.QtGui import QIcon

import FreeCAD
import FreeCADGui

from addonmanager_devmode_license_selector import LicenseSelector

translate = FreeCAD.Qt.translate

# pylint: disable=too-few-public-methods


class LicensesTable:
    """A QTableWidget and associated buttons for managing the list of authors and maintainers."""

    def __init__(self):
        self.widget = FreeCADGui.PySideUic.loadUi(
            os.path.join(os.path.dirname(__file__), "developer_mode_licenses_table.ui")
        )

        self.widget.addButton.setIcon(
            QIcon.fromTheme("add", QIcon(":/icons/list-add.svg"))
        )
        self.widget.removeButton.setIcon(
            QIcon.fromTheme("remove", QIcon(":/icons/list-remove.svg"))
        )

        self.widget.addButton.clicked.connect(self._add_clicked)
        self.widget.removeButton.clicked.connect(self._remove_clicked)
        self.widget.tableWidget.itemSelectionChanged.connect(self._selection_changed)
        self.widget.tableWidget.itemDoubleClicked.connect(self._edit)
        self.metadata = None
        self.path_to_addon = ""

    def show(self, metadata, path_to_addon):
        """Set up the widget based on incoming metadata"""
        self.metadata = metadata
        self.path_to_addon = path_to_addon
        self._populate_from_metadata()
        self.widget.removeButton.setDisabled(True)
        self.widget.show()

    def _populate_from_metadata(self):
        """Use the passed metadata object to populate the maintainers and authors"""
        self.widget.tableWidget.setRowCount(0)
        row = 0
        for lic in self.metadata.License:
            shortcode = lic["name"]
            path = lic["file"]
            self._add_row(row, shortcode, path)
            row += 1

    def _add_row(self, row, shortcode, path):
        """Add this license to the tableWidget at row given"""
        self.widget.tableWidget.insertRow(row)
        self.widget.tableWidget.setItem(row, 0, QTableWidgetItem(shortcode))
        self.widget.tableWidget.setItem(row, 1, QTableWidgetItem(path))

    def _add_clicked(self):
        """Callback: the Add License button was clicked"""
        dlg = LicenseSelector(self.path_to_addon)
        shortcode, path = dlg.exec()
        if shortcode and path:
            self._add_row(self.widget.tableWidget.rowCount(), shortcode, path)
            self.metadata.addLicense(shortcode, path)

    def _remove_clicked(self):
        """Callback: the Remove License button was clicked"""
        items = self.widget.tableWidget.selectedIndexes()
        if items:
            # We only support single-selection, so can just pull the row # from
            # the first entry
            row = items[0].row()
            shortcode = self.widget.tableWidget.item(row, 0).text()
            path = self.widget.tableWidget.item(row, 1).text()
            self.widget.tableWidget.removeRow(row)
            self.metadata.removeLicense(shortcode, path)

    def _edit(self, item):
        """Callback: a row in the tableWidget was double-clicked"""
        row = item.row()
        shortcode = self.widget.tableWidget.item(row, 0).text()
        path = self.widget.tableWidget.item(row, 1).text()

        dlg = LicenseSelector(self.path_to_addon)
        new_shortcode, new_path = dlg.exec(shortcode, path)

        if new_shortcode and new_path:
            self.widget.tableWidget.removeRow(row)
            self.metadata.removeLicense(new_shortcode, new_path)

            self._add_row(row, new_shortcode, new_path)
            self.metadata.addLicense(new_shortcode, new_path)

            self.widget.tableWidget.selectRow(row)

    def _selection_changed(self):
        """Callback: the current selection in the tableWidget changed"""
        items = self.widget.tableWidget.selectedItems()
        if items:
            self.widget.removeButton.setDisabled(False)
        else:
            self.widget.removeButton.setDisabled(True)



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2022 FreeCAD Project Association                        *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""GUI functions for uninstalling an Addon or Macro."""

import FreeCAD
import FreeCADGui

from PySide import QtCore, QtWidgets

from addonmanager_uninstaller import AddonUninstaller, MacroUninstaller
import addonmanager_utilities as utils

translate = FreeCAD.Qt.translate


class AddonUninstallerGUI(QtCore.QObject):
    """User interface for uninstalling an Addon: asks for confirmation, displays a progress dialog,
    displays completion and/or error dialogs, and emits the finished() signal when all work is
    complete."""

    finished = QtCore.Signal()

    def __init__(self, addon_to_remove):
        super().__init__()
        self.addon_to_remove = addon_to_remove
        if (
            hasattr(self.addon_to_remove, "macro")
            and self.addon_to_remove.macro is not None
        ):
            self.uninstaller = MacroUninstaller(self.addon_to_remove)
        else:
            self.uninstaller = AddonUninstaller(self.addon_to_remove)
        self.uninstaller.success.connect(self._succeeded)
        self.uninstaller.failure.connect(self._failed)
        self.worker_thread = QtCore.QThread()
        self.uninstaller.moveToThread(self.worker_thread)
        self.uninstaller.finished.connect(self.worker_thread.quit)
        self.worker_thread.started.connect(self.uninstaller.run)
        self.progress_dialog = None
        self.dialog_timer = QtCore.QTimer()
        self.dialog_timer.timeout.connect(self._show_progress_dialog)
        self.dialog_timer.setSingleShot(True)
        self.dialog_timer.setInterval(
            1000
        )  # Can override from external (e.g. testing) code

    def run(self):
        """Begin the user interaction: asynchronous, only blocks while showing the initial modal
        confirmation dialog."""
        ok_to_proceed = self._confirm_uninstallation()
        if not ok_to_proceed:
            self._finalize()
            return

        self.dialog_timer.start()
        self._run_uninstaller()

    def _confirm_uninstallation(self) -> bool:
        """Present a modal dialog asking the user if they really want to uninstall. Returns True to
        continue with the uninstallation, or False to stop the process."""
        confirm = QtWidgets.QMessageBox.question(
            utils.get_main_am_window(),
            translate("AddonsInstaller", "Confirm remove"),
            translate(
                "AddonsInstaller", "Are you sure you want to uninstall {}?"
            ).format(self.addon_to_remove.display_name),
            QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.Cancel,
        )
        return confirm == QtWidgets.QMessageBox.Yes

    def _show_progress_dialog(self):
        self.progress_dialog = QtWidgets.QMessageBox(
            QtWidgets.QMessageBox.NoIcon,
            translate("AddonsInstaller", "Removing Addon"),
            translate("AddonsInstaller", "Removing {}").format(
                self.addon_to_remove.display_name
            )
            + "...",
            QtWidgets.QMessageBox.Cancel,
            parent=utils.get_main_am_window(),
        )
        self.progress_dialog.rejected.connect(self._cancel_removal)
        self.progress_dialog.show()

    def _run_uninstaller(self):
        self.worker_thread.start()

    def _cancel_removal(self):
        """Ask the QThread to interrupt. Probably has no effect, most of the work is in a single OS
        call."""
        self.worker_thread.requestInterruption()

    def _succeeded(self, addon):
        """Callback for successful removal"""
        self.dialog_timer.stop()
        if self.progress_dialog:
            self.progress_dialog.hide()
        QtWidgets.QMessageBox.information(
            utils.get_main_am_window(),
            translate("AddonsInstaller", "Uninstall complete"),
            translate("AddonInstaller", "Finished removing {}").format(
                addon.display_name
            ),
        )
        self._finalize()

    def _failed(self, addon, message):
        """Callback for failed or partially failed removal"""
        self.dialog_timer.stop()
        if self.progress_dialog:
            self.progress_dialog.hide()
        QtWidgets.QMessageBox.critical(
            utils.get_main_am_window(),
            translate("AddonsInstaller", "Uninstall failed"),
            translate("AddonInstaller", "Failed to remove some files")
            + ":\n"
            + message,
        )
        self._finalize()

    def _finalize(self):
        """Clean up and emit finished signal"""
        if self.worker_thread.isRunning():
            self.worker_thread.requestInterruption()
            self.worker_thread.quit()
            self.worker_thread.wait(500)
        self.finished.emit()



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2022-2023 FreeCAD Project Association                   *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

""" Misc. worker thread classes for the FreeCAD Addon Manager. """

from typing import Optional

import FreeCAD
from PySide import QtCore
import addonmanager_utilities as utils

translate = FreeCAD.Qt.translate


class ConnectionChecker(QtCore.QThread):
    """A worker thread for checking the connection to GitHub as a proxy for overall
    network connectivity. It has two signals: success() and failure(str). The failure
    signal contains a translated error message suitable for display to an end user."""

    success = QtCore.Signal()
    failure = QtCore.Signal(str)

    def __init__(self):
        QtCore.QThread.__init__(self)

    def run(self):
        """Not generally called directly: create a new ConnectionChecker object and
        call start() on it to spawn a child thread."""

        FreeCAD.Console.PrintLog("Checking network connection...\n")
        result = self.check_network_connection()
        if QtCore.QThread.currentThread().isInterruptionRequested():
            FreeCAD.Console.PrintLog("Connection check cancelled\n")
            return
        if not result:
            self.failure.emit(
                translate(
                    "AddonsInstaller",
                    "Unable to read data from GitHub: check your internet connection and proxy settings and try again.",
                )
            )
            return
        FreeCAD.Console.PrintLog(f"GitHub's zen message response: {result}\n")
        self.success.emit()

    def check_network_connection(self) -> Optional[str]:
        """The main work of this object: returns the decoded result of the connection request, or
        None if the request failed"""
        url = "https://api.github.com/zen"
        result = utils.blocking_get(url)
        if result:
            return result.decode("utf8")
        return None



# ***************************************************************************
# *   Copyright (c) 2020 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

"""Provides GUI tools to set up default styles."""

## @package gui_setstyle
# \ingroup draftguitools
# \brief Provides GUI tools to set Draft styles such as color or line width

## \addtogroup draftguitools
# @{

import FreeCAD
import os
import draftutils.utils as utils

if FreeCAD.GuiUp:
    import FreeCADGui
    import Draft_rc
def QT_TRANSLATE_NOOP(ctx,txt):
    return txt
translate = FreeCAD.Qt.translate

__title__ = "FreeCAD Draft Workbench GUI Tools - Styling tools"
__author__ = ("Yorik van Havre")
__url__ = "https://www.freecad.org"

PRESETPATH = os.path.join(FreeCAD.getUserAppDataDir(),"Draft","StylePresets.json")


class Draft_SetStyle:

    """The Draft_SetStyle FreeCAD command definition."""

    def GetResources(self):

        d = {'Pixmap': 'Draft_Apply',
             'Accel': "S, S",
             'MenuText': QT_TRANSLATE_NOOP("Draft_SetStyle", "Set style"),
             'ToolTip': QT_TRANSLATE_NOOP("Draft_SetStyle", "Sets default styles")}
        return d

    def Activated(self):

        FreeCADGui.Control.showDialog(Draft_SetStyle_TaskPanel())



class Draft_SetStyle_TaskPanel:

    """The task panel for the Draft_SetStyle command"""

    def __init__(self):

        from PySide import QtCore,QtGui
        self.p = "User parameter:BaseApp/Preferences/"
        self.form = FreeCADGui.PySideUic.loadUi(":/ui/TaskPanel_SetStyle.ui")
        self.form.setWindowIcon(QtGui.QIcon.fromTheme("gtk-apply", QtGui.QIcon(":/icons/Draft_Apply.svg")))
        self.form.applyButton.setIcon(QtGui.QIcon.fromTheme("gtk-apply", QtGui.QIcon(":/icons/Draft_Apply.svg")))
        self.form.annotButton.setIcon(QtGui.QIcon(":/icons/Draft_Text.svg"))
        self.form.LineColor.setProperty("color",self.getPrefColor("View","DefaultShapeLineColor",255))
        self.form.LineWidth.setValue(FreeCAD.ParamGet(self.p+"View").GetInt("DefaultShapeLineWidth",2))
        self.form.DrawStyle.setCurrentIndex(FreeCAD.ParamGet(self.p+"Mod/Draft").GetInt("DefaultDrawStyle",0))
        self.form.DisplayMode.setCurrentIndex(FreeCAD.ParamGet(self.p+"Mod/Draft").GetInt("DefaultDisplayMode",0))
        self.form.ShapeColor.setProperty("color",self.getPrefColor("View","DefaultShapeColor",4294967295))
        self.form.Transparency.setValue(FreeCAD.ParamGet(self.p+"View").GetInt("DefaultShapeTransparency",0))
        self.form.TextFont.setCurrentFont(QtGui.QFont(FreeCAD.ParamGet(self.p+"Mod/Draft").GetString("textfont","Sans")))
        self.form.TextSize.setText(FreeCAD.Units.Quantity(FreeCAD.ParamGet(self.p+"Mod/Draft").GetFloat("textheight",10),FreeCAD.Units.Length).UserString)
        self.form.TextColor.setProperty("color",self.getPrefColor("Mod/Draft","DefaultTextColor",255))
        self.form.ArrowStyle.setCurrentIndex(FreeCAD.ParamGet(self.p+"Mod/Draft").GetInt("dimsymbol",0))
        self.form.ArrowSize.setText(FreeCAD.Units.Quantity(FreeCAD.ParamGet(self.p+"Mod/Draft").GetFloat("arrowsize",5),FreeCAD.Units.Length).UserString)
        self.form.ShowUnit.setChecked(FreeCAD.ParamGet(self.p+"Mod/Draft").GetBool("showUnit",True))
        self.form.UnitOverride.setText(FreeCAD.ParamGet(self.p+"Mod/Draft").GetString("overrideUnit",""))
        self.form.saveButton.setIcon(QtGui.QIcon.fromTheme("gtk-save", QtGui.QIcon(":/icons/document-save.svg")))
        self.form.TextSpacing.setText(FreeCAD.Units.Quantity(FreeCAD.ParamGet(self.p+"Mod/Draft").GetFloat("dimspacing",1),FreeCAD.Units.Length).UserString)
        self.form.LineSpacing.setValue(FreeCAD.ParamGet(self.p+"Mod/Draft").GetFloat("LineSpacing",1))
        self.form.DimOvershoot.setText(FreeCAD.Units.Quantity(FreeCAD.ParamGet(self.p+"Mod/Draft").GetFloat("dimovershoot",0),FreeCAD.Units.Length).UserString)
        self.form.ExtLines.setText(FreeCAD.Units.Quantity(FreeCAD.ParamGet(self.p+"Mod/Draft").GetFloat("extlines",0),FreeCAD.Units.Length).UserString)
        self.form.ExtOvershoot.setText(FreeCAD.Units.Quantity(FreeCAD.ParamGet(self.p+"Mod/Draft").GetFloat("extovershoot",0),FreeCAD.Units.Length).UserString)
        self.form.saveButton.clicked.connect(self.onSaveStyle)
        self.form.applyButton.clicked.connect(self.onApplyStyle)
        self.form.annotButton.clicked.connect(self.onApplyAnnot)
        self.form.comboPresets.currentIndexChanged.connect(self.onLoadStyle)
        self.loadDefaults()

    def loadDefaults(self):

        presets = [self.form.comboPresets.itemText(0)]
        self.form.comboPresets.clear()
        pdict = self.load()
        presets.extend(pdict.keys())
        self.form.comboPresets.addItems(presets)
        current = self.getValues()
        for name,preset in pdict.items():
            if all(item in current.items() for item in preset.items()): #if preset == current:
                self.form.comboPresets.setCurrentIndex(1+(list(pdict.keys()).index(name)))
                break

    def getPrefColor(self,group,prop,default):

        c = FreeCAD.ParamGet(self.p+group).GetUnsigned(prop,default)
        return self.getColor(c)

    def getColor(self,c):

        from PySide import QtGui
        return QtGui.QColor(utils.rgba_to_argb(c))

    def getValues(self):

        preset = {}
        preset["LineColor"] = utils.argb_to_rgba(self.form.LineColor.property("color").rgba())
        preset["LineWidth"] = self.form.LineWidth.value()
        preset["DrawStyle"] = self.form.DrawStyle.currentIndex()
        preset["DisplayMode"] = self.form.DisplayMode.currentIndex()
        preset["ShapeColor"] = utils.argb_to_rgba(self.form.ShapeColor.property("color").rgba())
        preset["Transparency"] = self.form.Transparency.value()
        preset["TextFont"] = self.form.TextFont.currentFont().family()
        preset["TextSize"] = FreeCAD.Units.Quantity(self.form.TextSize.text()).Value
        preset["TextColor"] = utils.argb_to_rgba(self.form.TextColor.property("color").rgba())
        preset["ArrowStyle"] = self.form.ArrowStyle.currentIndex()
        preset["ArrowSize"] = FreeCAD.Units.Quantity(self.form.ArrowSize.text()).Value
        preset["ShowUnit"] = self.form.ShowUnit.isChecked()
        preset["UnitOverride"] = self.form.UnitOverride.text()
        preset["TextSpacing"] = FreeCAD.Units.Quantity(self.form.TextSpacing.text()).Value
        preset["LineSpacing"] = self.form.LineSpacing.value()
        preset["DimOvershoot"] = FreeCAD.Units.Quantity(self.form.DimOvershoot.text()).Value
        preset["ExtLines"] = FreeCAD.Units.Quantity(self.form.ExtLines.text()).Value
        preset["ExtOvershoot"] = FreeCAD.Units.Quantity(self.form.ExtOvershoot.text()).Value
        return preset

    def setValues(self,preset):

        from PySide import QtCore,QtGui
        self.form.LineColor.setProperty("color",self.getColor(preset.get("LineColor",255)))
        self.form.LineWidth.setValue(preset.get("LineWidth",1))
        self.form.DrawStyle.setCurrentIndex(preset.get("DrawStyle",0))
        self.form.DisplayMode.setCurrentIndex(preset.get("DisplayMode",0))
        self.form.ShapeColor.setProperty("color",self.getColor(preset.get("ShapeColor",4294967295)))
        self.form.Transparency.setValue(preset.get("Transparency",0))
        self.form.TextFont.setCurrentFont(QtGui.QFont(preset.get("TextFont","sans")))
        self.form.TextColor.setProperty("color",self.getColor(preset.get("TextColor",255)))
        self.form.ArrowStyle.setCurrentIndex(preset.get("ArrowStyle",0))
        self.form.ShowUnit.setChecked(preset.get("ShowUnit",False))
        self.form.UnitOverride.setText(preset.get("UnitOverride",""))
        self.form.TextSize.setText(FreeCAD.Units.Quantity(preset.get("TextSize",25),FreeCAD.Units.Length).UserString)
        self.form.ArrowSize.setText(FreeCAD.Units.Quantity(preset.get("ArrowSize",5),FreeCAD.Units.Length).UserString)
        self.form.TextSpacing.setText(FreeCAD.Units.Quantity(preset.get("TextSpacing",25),FreeCAD.Units.Length).UserString)
        self.form.LineSpacing.setValue(preset.get("LineSpacing",3))
        self.form.DimOvershoot.setText(FreeCAD.Units.Quantity(preset.get("DimOvershoot",0),FreeCAD.Units.Length).UserString)
        self.form.ExtLines.setText(FreeCAD.Units.Quantity(preset.get("ExtLines",0),FreeCAD.Units.Length).UserString)
        self.form.ExtOvershoot.setText(FreeCAD.Units.Quantity(preset.get("ExtOvershoot",0),FreeCAD.Units.Length).UserString)

    def reject(self):

        FreeCADGui.Control.closeDialog()

    def accept(self):

        param_draft = FreeCAD.ParamGet(self.p + "Mod/Draft")
        param_view  = FreeCAD.ParamGet(self.p + "View")

        param_view.SetUnsigned("DefaultShapeLineColor",utils.argb_to_rgba(self.form.LineColor.property("color").rgba()))
        param_view.SetInt("DefaultShapeLineWidth",self.form.LineWidth.value())
        param_view.SetUnsigned("DefaultShapeVertexColor",utils.argb_to_rgba(self.form.LineColor.property("color").rgba()))
        param_view.SetInt("DefaultShapePointSize",self.form.LineWidth.value())
        param_draft.SetInt("DefaultDrawStyle",self.form.DrawStyle.currentIndex())
        param_draft.SetInt("DefaultDisplayMode",self.form.DisplayMode.currentIndex())
        param_view.SetUnsigned("DefaultShapeColor",utils.argb_to_rgba(self.form.ShapeColor.property("color").rgba()))
        param_view.SetInt("DefaultShapeTransparency",self.form.Transparency.value())
        param_draft.SetString("textfont",self.form.TextFont.currentFont().family())
        param_draft.SetFloat("textheight",FreeCAD.Units.Quantity(self.form.TextSize.text()).Value)
        param_draft.SetUnsigned("DefaultTextColor",utils.argb_to_rgba(self.form.TextColor.property("color").rgba()))
        param_draft.SetInt("dimsymbol",self.form.ArrowStyle.currentIndex())
        param_draft.SetFloat("arrowsize",FreeCAD.Units.Quantity(self.form.ArrowSize.text()).Value)
        param_draft.SetBool("showUnit",self.form.ShowUnit.isChecked())
        param_draft.SetString("overrideUnit",self.form.UnitOverride.text())
        param_draft.SetFloat("dimspacing",FreeCAD.Units.Quantity(self.form.TextSpacing.text()).Value)
        param_draft.SetFloat("LineSpacing",self.form.LineSpacing.value())
        param_draft.SetFloat("dimovershoot",FreeCAD.Units.Quantity(self.form.DimOvershoot.text()).Value)
        param_draft.SetFloat("extlines",FreeCAD.Units.Quantity(self.form.ExtLines.text()).Value)
        param_draft.SetFloat("extovershoot",FreeCAD.Units.Quantity(self.form.ExtOvershoot.text()).Value)
        if hasattr(FreeCADGui,"draftToolBar"):
            FreeCADGui.draftToolBar.setStyleButton()
        self.reject()

    def onApplyStyle(self):

        for obj in FreeCADGui.Selection.getSelection():
            self.apply_style_to_obj(obj)

    def onApplyAnnot(self):

        if FreeCAD.ActiveDocument is not None: # Command can be called without a document.
            from draftutils import utils
            objs = FreeCAD.ActiveDocument.Objects
            typs = ["Dimension", "LinearDimension", "AngularDimension",
                    "Text", "DraftText", "Label"]
            for obj in objs:
                if utils.get_type(obj) in typs:
                    self.apply_style_to_obj(obj)

    def apply_style_to_obj(self, obj):

        vobj = obj.ViewObject
        if not vobj:
            return

        properties = vobj.PropertiesList
        if "LineColor" in properties:
            vobj.LineColor = self.form.LineColor.property("color").getRgbF()[:3] # Remove Alpha (which is 1 instead of 0).
        if "LineWidth" in properties:
            vobj.LineWidth = self.form.LineWidth.value()
        if "PointColor" in properties:
            vobj.PointColor = self.form.LineColor.property("color").getRgbF()[:3]
        if "PointSize" in properties:
            vobj.PointSize = self.form.LineWidth.value()
        if "DrawStyle" in properties:
            vobj.DrawStyle = ["Solid", "Dashed", "Dotted", "Dashdot"][self.form.DrawStyle.currentIndex()]
        if "DisplayMode" in properties:
            dmodes = ["Flat Lines", "Wireframe", "Shaded", "Points"]
            dm = dmodes[self.form.DisplayMode.currentIndex()]
            if dm in vobj.getEnumerationsOfProperty("DisplayMode"):
                vobj.DisplayMode = dm
        if "ShapeColor" in properties:
            vobj.ShapeColor = self.form.ShapeColor.property("color").getRgbF()[:3]
        if "Transparency" in properties:
            vobj.Transparency = self.form.Transparency.value()
        if "FontName" in properties:
            vobj.FontName = self.form.TextFont.currentFont().family()
        if "FontSize" in properties:
            vobj.FontSize = FreeCAD.Units.Quantity(self.form.TextSize.text()).Value
        if "TextColor" in properties:
            vobj.TextColor = self.form.TextColor.property("color").getRgbF()[:3]
        if "ArrowType" in properties:
            vobj.ArrowType = ["Dot", "Circle", "Arrow", "Tick", "Tick-2"][self.form.ArrowStyle.currentIndex()]
        if "ArrowSize" in properties:
            vobj.ArrowSize = FreeCAD.Units.Quantity(self.form.ArrowSize.text()).Value
        if "ShowUnit" in properties:
            vobj.ShowUnit = self.form.ShowUnit.isChecked()
        if "UnitOverride" in properties:
            vobj.UnitOverride = self.form.UnitOverride.text()
        if "TextSpacing" in properties:
            vobj.TextSpacing = FreeCAD.Units.Quantity(self.form.TextSpacing.text()).Value
        if "LineSpacing" in properties:
            vobj.LineSpacing = self.form.LineSpacing.value()
        if "DimOvershoot" in properties:
            vobj.DimOvershoot = FreeCAD.Units.Quantity(self.form.DimOvershoot.text()).Value
        if "ExtLines" in properties:
            vobj.ExtLines = FreeCAD.Units.Quantity(self.form.ExtLines.text()).Value
        if "ExtOvershoot" in properties:
            vobj.ExtOvershoot = FreeCAD.Units.Quantity(self.form.ExtOvershoot.text()).Value

    def onLoadStyle(self,index):

        if index > 0:
            pdict = self.load()
            if self.form.comboPresets.itemText(index) in pdict.keys():
                preset = pdict[self.form.comboPresets.itemText(index)]
                self.setValues(preset)

    def onSaveStyle(self):

        from PySide import QtCore,QtGui
        reply = QtGui.QInputDialog.getText(None,
                                           translate("Draft","Save style"),
                                           translate("Draft","Name of this new style:"))
        if reply[1]:
            name = reply[0]
            pdict = self.load()
            if pdict:
                if name in pdict.keys():
                    reply = QtGui.QMessageBox.question(None,
                                                       translate("Draft","Warning"),
                                                       translate("Draft","Name exists. Overwrite?"),
                                                       QtGui.QMessageBox.Yes | QtGui.QMessageBox.No,
                                                       QtGui.QMessageBox.No)
                    if reply == QtGui.QMessageBox.No:
                        return
            preset = self.getValues()
            pdict[name] = preset
            self.save(pdict)
            self.loadDefaults()

    def load(self):

        """loads the presets json file, returns a dict"""

        pdict = {}
        try:
            import json
            from json.decoder import JSONDecodeError
        except Exception:
            FreeCAD.Console.PrintError(translate("Draft","Error: json module not found. Unable to load style")+"\n")
            return
        if os.path.exists(PRESETPATH):
            with open(PRESETPATH,"r") as f:
                try:
                    pdict = json.load(f)
                except JSONDecodeError:
                    return {}
        return pdict

    def save(self,d):

        """saves the given dict to the presets json file"""

        try:
            import json
        except Exception:
            FreeCAD.Console.PrintError(translate("Draft","Error: json module not found. Unable to save style")+"\n")
            return
        folder = os.path.dirname(PRESETPATH)
        if not os.path.exists(folder):
            os.makedirs(folder)
        with open(PRESETPATH,"w") as f:
            json.dump(d,f,indent=4)


FreeCADGui.addCommand('Draft_SetStyle', Draft_SetStyle())

## @}



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2022-2023 FreeCAD Project Association                   *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""Mock objects for use when testing the addon manager non-GUI code."""

# pylint: disable=too-few-public-methods,too-many-instance-attributes,missing-function-docstring

import os
from typing import Union, List
import xml.etree.ElementTree as ElemTree


class GitFailed(RuntimeError):
    pass


class MockConsole:
    """Mock for the FreeCAD.Console -- does NOT print anything out, just logs it."""

    def __init__(self):
        self.log = []
        self.messages = []
        self.warnings = []
        self.errors = []

    def PrintLog(self, data: str):
        self.log.append(data)

    def PrintMessage(self, data: str):
        self.messages.append(data)

    def PrintWarning(self, data: str):
        self.warnings.append(data)

    def PrintError(self, data: str):
        self.errors.append(data)

    def missing_newlines(self) -> int:
        """In most cases, all console entries should end with newlines: this is a
        convenience function for unit testing that is true."""
        counter = 0
        counter += self._count_missing_newlines(self.log)
        counter += self._count_missing_newlines(self.messages)
        counter += self._count_missing_newlines(self.warnings)
        counter += self._count_missing_newlines(self.errors)
        return counter

    @staticmethod
    def _count_missing_newlines(some_list) -> int:
        counter = 0
        for line in some_list:
            if line[-1] != "\n":
                counter += 1
        return counter


class MockAddon:
    """Minimal Addon class"""

    # pylint: disable=too-many-instance-attributes

    def __init__(
        self,
        name: str = None,
        url: str = None,
        status: object = None,
        branch: str = "main",
    ):
        test_dir = os.path.join(os.path.dirname(__file__), "..", "data")
        if name:
            self.name = name
            self.display_name = name
        else:
            self.name = "MockAddon"
            self.display_name = "Mock Addon"
        self.url = url if url else os.path.join(test_dir, "test_simple_repo.zip")
        self.branch = branch
        self.status = status
        self.macro = None
        self.update_status = None
        self.metadata = None
        self.icon_file = None
        self.last_updated = None
        self.requires = set()
        self.python_requires = set()
        self.python_optional = set()
        self.on_git = False
        self.on_wiki = True

    def set_status(self, status):
        self.update_status = status

    @staticmethod
    def get_best_icon_relative_path():
        return ""


class MockMacro:
    """Minimal Macro class"""

    def __init__(self, name="MockMacro"):
        self.name = name
        self.filename = self.name + ".FCMacro"
        self.icon = ""  # If set, should just be fake filename, doesn't have to exist
        self.xpm = ""
        self.code = ""
        self.raw_code_url = ""
        self.other_files = []  # If set, should be fake names, don't have to exist
        self.details_filled_from_file = False
        self.details_filled_from_code = False
        self.parsed_wiki_page = False
        self.on_git = False
        self.on_wiki = True

    def install(self, location: os.PathLike):
        """Installer function for the mock macro object: creates a file with the src_filename
        attribute, and optionally an icon, xpm, and other_files. The data contained in these files
        is not usable and serves only as a placeholder for the existence of the files.
        """

        with open(
            os.path.join(location, self.filename),
            "w",
            encoding="utf-8",
        ) as f:
            f.write("Test file for macro installation unit tests")
        if self.icon:
            with open(os.path.join(location, self.icon), "wb") as f:
                f.write(b"Fake icon data - nothing to see here\n")
        if self.xpm:
            with open(
                os.path.join(location, "MockMacro_icon.xpm"), "w", encoding="utf-8"
            ) as f:
                f.write(self.xpm)
        for name in self.other_files:
            if "/" in name:
                new_location = os.path.dirname(os.path.join(location, name))
                os.makedirs(new_location, exist_ok=True)
            with open(os.path.join(location, name), "w", encoding="utf-8") as f:
                f.write("# Fake macro data for unit testing\n")
        return True, []

    def fill_details_from_file(self, _):
        """Tracks that this function was called, but otherwise does nothing"""
        self.details_filled_from_file = True

    def fill_details_from_code(self, _):
        self.details_filled_from_code = True

    def parse_wiki_page(self, _):
        self.parsed_wiki_page = True


class SignalCatcher:
    """Object to track signals that it has caught.

    Usage:
    catcher = SignalCatcher()
    my_signal.connect(catcher.catch_signal)
    do_things_that_emit_the_signal()
    self.assertTrue(catcher.caught)
    """

    def __init__(self):
        self.caught = False
        self.killed = False
        self.args = None

    def catch_signal(self, *args):
        self.caught = True
        self.args = args

    def die(self):
        self.killed = True


class AddonSignalCatcher:
    """Signal catcher specifically designed for catching emitted addons."""

    def __init__(self):
        self.addons = []

    def catch_signal(self, addon):
        self.addons.append(addon)


class CallCatcher:
    """Generic call monitor -- use to override functions that are not themselves under
    test so that you can detect when the function has been called, and how many times.
    """

    def __init__(self):
        self.called = False
        self.call_count = 0
        self.args = None

    def catch_call(self, *args):
        self.called = True
        self.call_count += 1
        self.args = args


class MockGitManager:
    """A mock git manager: does NOT require a git installation. Takes no actions, only records
    which functions are called for instrumentation purposes. Can be forced to appear to fail as
    needed. Various member variables can be set to emulate necessary return responses.
    """

    def __init__(self):
        self.called_methods = []
        self.update_available_response = False
        self.current_tag_response = "main"
        self.current_branch_response = "main"
        self.get_remote_response = "No remote set"
        self.get_branches_response = ["main"]
        self.get_last_committers_response = {
            "John Doe": {"email": "jdoe@freecad.org", "count": 1}
        }
        self.get_last_authors_response = {
            "Jane Doe": {"email": "jdoe@freecad.org", "count": 1}
        }
        self.should_fail = False
        self.fail_once = False  # Switch back to success after the simulated failure

    def _check_for_failure(self):
        if self.should_fail:
            if self.fail_once:
                self.should_fail = False
            raise GitFailed("Unit test forced failure")

    def clone(self, _remote, _local_path, _args: List[str] = None):
        self.called_methods.append("clone")
        self._check_for_failure()

    def async_clone(
        self, _remote, _local_path, _progress_monitor, _args: List[str] = None
    ):
        self.called_methods.append("async_clone")
        self._check_for_failure()

    def checkout(self, _local_path, _spec, _args: List[str] = None):
        self.called_methods.append("checkout")
        self._check_for_failure()

    def update(self, _local_path):
        self.called_methods.append("update")
        self._check_for_failure()

    def status(self, _local_path) -> str:
        self.called_methods.append("status")
        self._check_for_failure()
        return "Up-to-date"

    def reset(self, _local_path, _args: List[str] = None):
        self.called_methods.append("reset")
        self._check_for_failure()

    def async_fetch_and_update(self, _local_path, _progress_monitor, _args=None):
        self.called_methods.append("async_fetch_and_update")
        self._check_for_failure()

    def update_available(self, _local_path) -> bool:
        self.called_methods.append("update_available")
        self._check_for_failure()
        return self.update_available_response

    def current_tag(self, _local_path) -> str:
        self.called_methods.append("current_tag")
        self._check_for_failure()
        return self.current_tag_response

    def current_branch(self, _local_path) -> str:
        self.called_methods.append("current_branch")
        self._check_for_failure()
        return self.current_branch_response

    def repair(self, _remote, _local_path):
        self.called_methods.append("repair")
        self._check_for_failure()

    def get_remote(self, _local_path) -> str:
        self.called_methods.append("get_remote")
        self._check_for_failure()
        return self.get_remote_response

    def get_branches(self, _local_path) -> List[str]:
        self.called_methods.append("get_branches")
        self._check_for_failure()
        return self.get_branches_response

    def get_last_committers(self, _local_path, _n=10):
        self.called_methods.append("get_last_committers")
        self._check_for_failure()
        return self.get_last_committers_response

    def get_last_authors(self, _local_path, _n=10):
        self.called_methods.append("get_last_authors")
        self._check_for_failure()
        return self.get_last_authors_response


class MockSignal:
    """A purely synchronous signal, instrumented and intended only for use in unit testing.
    emit() is semi-functional, but does not use queued slots so cannot be used across
    threads."""

    def __init__(self, *args):
        self.expected_types = args
        self.connections = []
        self.disconnections = []
        self.emitted = False

    def connect(self, func):
        self.connections.append(func)

    def disconnect(self, func):
        if func in self.connections:
            self.connections.remove(func)
        self.disconnections.append(func)

    def emit(self, *args):
        self.emitted = True
        for connection in self.connections:
            connection(args)


class MockNetworkManager:
    """Instrumented mock for the NetworkManager. Does no network access, is not asynchronous, and
    does not require a running event loop. No submitted requests ever complete."""

    def __init__(self):
        self.urls = []
        self.aborted = []
        self.data = MockByteArray()
        self.called_methods = []

        self.completed = MockSignal(int, int, MockByteArray)
        self.progress_made = MockSignal(int, int, int)
        self.progress_complete = MockSignal(int, int, os.PathLike)

    def submit_unmonitored_get(self, url: str) -> int:
        self.urls.append(url)
        self.called_methods.append("submit_unmonitored_get")
        return len(self.urls) - 1

    def submit_monitored_get(self, url: str) -> int:
        self.urls.append(url)
        self.called_methods.append("submit_monitored_get")
        return len(self.urls) - 1

    def blocking_get(self, url: str):
        self.urls.append(url)
        self.called_methods.append("blocking_get")
        return self.data

    def abort_all(self):
        self.called_methods.append("abort_all")
        for url in self.urls:
            self.aborted.append(url)

    def abort(self, index: int):
        self.called_methods.append("abort")
        self.aborted.append(self.urls[index])


class MockByteArray:
    """Mock for QByteArray. Only provides the data() access member."""

    def __init__(self, data_to_wrap="data".encode("utf-8")):
        self.wrapped = data_to_wrap

    def data(self) -> bytes:
        return self.wrapped


class MockThread:
    """Mock for QThread for use when threading is not being used, but interruption
    needs to be tested. Set interrupt_after_n_calls to the call number to stop at."""

    def __init__(self):
        self.interrupt_after_n_calls = 0
        self.interrupt_check_counter = 0

    def isInterruptionRequested(self):
        self.interrupt_check_counter += 1
        if (
            self.interrupt_after_n_calls
            and self.interrupt_check_counter >= self.interrupt_after_n_calls
        ):
            return True
        return False



# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides various functions to work with faces."""
## @package faces
# \ingroup draftgeoutils
# \brief Provides various functions to work with faces.

import lazy_loader.lazy_loader as lz

import DraftVecUtils
from FreeCAD import Base
from draftgeoutils.general import precision
from draftgeoutils.geometry import are_coplanar

# Delay import of module until first use because it is heavy
Part = lz.LazyLoader("Part", globals(), "Part")

## \addtogroup draftgeoutils
# @{


def concatenate(shape):
    """Turn several faces into one."""
    boundary_edges = getBoundary(shape)
    sorted_edges = Part.sortEdges(boundary_edges)

    try:
        wires = [Part.Wire(edges) for edges in sorted_edges]
        face = Part.makeFace(wires, "Part::FaceMakerBullseye")
    except Base.FreeCADError:
        print("DraftGeomUtils: Fails to join faces into one. "
              + "The precision of the faces would be insufficient")
        return shape
    else:
        if not wires[0].isClosed():
            return wires[0]
        else:
            return face


def getBoundary(shape):
    """Return the boundary edges of a group of faces."""
    if isinstance(shape, list):
        shape = Part.makeCompound(shape)

    # Make a lookup-table where we get the number of occurrences
    # to each edge in the fused face
    table = dict()
    for f in shape.Faces:
        for e in f.Edges:
            hash_code = e.hashCode()
            if hash_code in table:
                table[hash_code] = table[hash_code] + 1
            else:
                table[hash_code] = 1

    # Filter out the edges shared by more than one sub-face
    bound = list()
    for e in shape.Edges:
        if table[e.hashCode()] == 1:
            bound.append(e)
    return bound


def is_coplanar(faces, tol=-1):
    """Return True if all faces in the given list are coplanar.

    Parameters
    ----------
    faces: list
        List of faces to check coplanarity.
    tol: float, optional
        It defaults to `-1`, the tolerance of confusion, equal to 1e-7.
        Is the maximum deviation to be considered coplanar.

    Returns
    -------
    out: bool
        True if all face are coplanar. False in other case.
    """

    first_face = faces[0]
    for face in faces:
        if not are_coplanar(first_face, face, tol):
            return False

    return True

isCoplanar = is_coplanar


def bind(w1, w2, per_segment=False):
    """Bind 2 wires by their endpoints and returns a face.

    If per_segment is True and the wires have the same number of edges, the
    wires are processed per segment: a separate face is created for each pair
    of edges (one from w1 and one from w2), and the faces are then fused. This
    avoids problems with walls based on wires that selfintersect, or that have
    a loop that ends in a T-connection (f.e. a wire shaped like a number 6).
    """

    def create_face(w1, w2):
        try:
            w3 = Part.LineSegment(w1.Vertexes[0].Point,
                                  w2.Vertexes[0].Point).toShape()
            w4 = Part.LineSegment(w1.Vertexes[-1].Point,
                                  w2.Vertexes[-1].Point).toShape()
            return Part.Face(Part.Wire(w1.Edges + [w3] + w2.Edges + [w4]))
        except Part.OCCError:
            print("DraftGeomUtils: unable to bind wires")
            return None

    if not w1 or not w2:
        print("DraftGeomUtils: unable to bind wires")
        return None

    if (per_segment
            and len(w1.Edges) > 1
            and len(w1.Edges) == len(w2.Edges)):
        faces = []
        for (edge1, edge2) in zip(w1.Edges, w2.Edges):
            face = create_face(edge1, edge2)
            if face is None:
                return None
            faces.append(face)
        # return concatenate(faces[0].fuse(faces[1:])) # Also works.
        return faces[0].fuse(faces[1:]).removeSplitter().Faces[0]
    elif w1.isClosed() and w2.isClosed():
        d1 = w1.BoundBox.DiagonalLength
        d2 = w2.BoundBox.DiagonalLength
        if d1 < d2:
            w1, w2 = w2, w1
        # return Part.Face(w1).cut(Part.Face(w2)).Faces[0] # Only works if wires do not self-intersect.
        try:
            face = Part.Face([w1, w2])
            face.fix(1e-7, 0, 1)
            return face
        except Part.OCCError:
            print("DraftGeomUtils: unable to bind wires")
            return None
    else:
        return create_face(w1, w2)


def cleanFaces(shape):
    """Remove inner edges from coplanar faces."""
    faceset = shape.Faces

    def find(hc):
        """Find a face with the given hashcode."""
        for f in faceset:
            if f.hashCode() == hc:
                return f

    def findNeighbour(hface, hfacelist):
        """Find the first neighbour of a face, and return its index."""
        eset = []
        for e in find(hface).Edges:
            eset.append(e.hashCode())
        for i in range(len(hfacelist)):
            for ee in find(hfacelist[i]).Edges:
                if ee.hashCode() in eset:
                    return i
        return None

    # build lookup table
    lut = {}
    for face in faceset:
        for edge in face.Edges:
            if edge.hashCode() in lut:
                lut[edge.hashCode()].append(face.hashCode())
            else:
                lut[edge.hashCode()] = [face.hashCode()]

    # print("lut:",lut)
    # take edges shared by 2 faces
    sharedhedges = []
    for k, v in lut.items():
        if len(v) == 2:
            sharedhedges.append(k)

    # print(len(sharedhedges)," shared edges:",sharedhedges)
    # find those with same normals
    targethedges = []
    for hedge in sharedhedges:
        faces = lut[hedge]
        n1 = find(faces[0]).normalAt(0.5, 0.5)
        n2 = find(faces[1]).normalAt(0.5, 0.5)
        if n1 == n2:
            targethedges.append(hedge)

    # print(len(targethedges)," target edges:",targethedges)
    # get target faces
    hfaces = []
    for hedge in targethedges:
        for f in lut[hedge]:
            if f not in hfaces:
                hfaces.append(f)

    # print(len(hfaces)," target faces:",hfaces)
    # sort islands
    islands = [[hfaces.pop(0)]]
    currentisle = 0
    currentface = 0
    found = True
    while hfaces:
        if not found:
            if len(islands[currentisle]) > (currentface + 1):
                currentface += 1
                found = True
            else:
                islands.append([hfaces.pop(0)])
                currentisle += 1
                currentface = 0
                found = True
        else:
            f = findNeighbour(islands[currentisle][currentface], hfaces)
            if f is not None:
                islands[currentisle].append(hfaces.pop(f))
            else:
                found = False

    # print(len(islands)," islands:",islands)
    # make new faces from islands
    newfaces = []
    treated = []
    for isle in islands:
        treated.extend(isle)
        fset = []
        for i in isle:
            fset.append(find(i))
        bounds = getBoundary(fset)
        shp = Part.Wire(Part.__sortEdges__(bounds))
        shp = Part.Face(shp)
        if shp.normalAt(0.5, 0.5) != find(isle[0]).normalAt(0.5, 0.5):
            shp.reverse()
        newfaces.append(shp)

    # print("new faces:",newfaces)
    # add remaining faces
    for f in faceset:
        if not f.hashCode() in treated:
            newfaces.append(f)

    # print("final faces")
    # finishing
    fshape = Part.makeShell(newfaces)
    if shape.isClosed():
        fshape = Part.makeSolid(fshape)
    return fshape


def removeSplitter(shape):
    """Return a face from removing the splitter in a list of faces.

    This is an alternative, shared edge-based version of Part.removeSplitter.
    Returns a face, or `None` if the operation failed.
    """
    lookup = dict()
    for f in shape.Faces:
        for e in f.Edges:
            h = e.hashCode()
            if h in lookup:
                lookup[h].append(e)
            else:
                lookup[h] = [e]

    edges = [e[0] for e in lookup.values() if len(e) == 1]

    try:
        face = Part.Face(Part.Wire(edges))
    except Part.OCCError:
        # operation failed
        return None
    else:
        if face.isValid():
            return face

    return None

## @}



# ***************************************************************************
# *   Copyright (c) 2023 Uwe Stöhr <uwestoehr@lyx.org>                      *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM Magnetodynamics2D Elmer writer"
__author__ = "Uwe Stöhr"
__url__ = "https://www.freecad.org"

## \addtogroup FEM
#  @{

from FreeCAD import Console
from FreeCAD import Units

from .. import sifio
from .. import writer as general_writer


class MgDyn2Dwriter:

    def __init__(self, writer, solver):
        self.write = writer
        self.solver = solver

    def getMagnetodynamic2DSolver(self, equation):
        # output the equation parameters
        s = self.write.createNonlinearSolver(equation)
        if not equation.IsHarmonic:
            s["Equation"] = "MgDyn2D"
            s["Procedure"] = sifio.FileAttr("MagnetoDynamics2D/MagnetoDynamics2D")
            s["Variable"] = "Potential"
        else:
            s["Equation"] = "MgDyn2DHarmonic"
            s["Procedure"] = sifio.FileAttr("MagnetoDynamics2D/MagnetoDynamics2DHarmonic")
            s["Variable"] = "Potential[Potential Re:1 Potential Im:1]"
        s["Exec Solver"] = "Always"
        s["Optimize Bandwidth"] = True
        s["Stabilize"] = equation.Stabilize
        return s

    def getMagnetodynamic2DSolverPost(self, equation):
        # output the equation parameters
        s = self.write.createNonlinearSolver(equation)
        s["Equation"] = "MgDyn2DPost"
        s["Exec Solver"] = "Always"
        s["Procedure"] = sifio.FileAttr("MagnetoDynamics/MagnetoDynamicsCalcFields")
        if equation.IsHarmonic:
            s["Angular Frequency"] = float(Units.Quantity(equation.AngularFrequency).Value)
        s["Potential Variable"] = "Potential"
        if equation.CalculateCurrentDensity is True:
            s["Calculate Current Density"] = True
        if equation.CalculateElectricField is True:
            s["Calculate Electric Field"] = True
        if equation.CalculateElementalFields is False:
            s["Calculate Elemental Fields"] = False
        if equation.CalculateHarmonicLoss is True:
            s["Calculate Harmonic Loss"] = True
        if equation.CalculateJouleHeating is True:
            s["Calculate Joule Heating"] = True
        if equation.CalculateMagneticFieldStrength is True:
            s["Calculate Magnetic Field Strength"] = True
        if equation.CalculateMaxwellStress is True:
            s["Calculate Maxwell Stress"] = True
        if equation.CalculateNodalFields is False:
            s["Calculate Nodal Fields"] = False
        if equation.CalculateNodalForces is True:
            s["Calculate Nodal Forces"] = True
        if equation.CalculateNodalHeating is True:
            s["Calculate Nodal Heating"] = True
        s["Optimize Bandwidth"] = True
        s["Stabilize"] = equation.Stabilize
        return s

    def handleMagnetodynamic2DConstants(self):
        permeability = self.write.convert(
            self.write.constsdef["PermeabilityOfVacuum"],
            "M*L/(T^2*I^2)"
        )
        # we round in the following to get rid of numerical artifacts
        self.write.constant("Permeability Of Vacuum", round(permeability, 20))

        permittivity = self.write.convert(
            self.write.constsdef["PermittivityOfVacuum"],
            "T^4*I^2/(L^3*M)"
        )
        self.write.constant("Permittivity Of Vacuum", round(permittivity, 20))

    def handleMagnetodynamic2DMaterial(self, bodies):
        # check that all bodies have a set material
        for name in bodies:
            if self.write.getBodyMaterial(name) is None:
                raise general_writer.WriteError(
                    "The body {} is not referenced in any material.\n\n".format(name)
                )
        for obj in self.write.getMember("App::MaterialObject"):
            m = obj.Material
            refs = (
                obj.References[0][1]
                if obj.References
                else self.write.getAllBodies())
            for name in (n for n in refs if n in bodies):
                if "ElectricalConductivity" not in m:
                    Console.PrintMessage("m: {}\n".format(m))
                    raise general_writer.WriteError(
                        "The electrical conductivity must be specified for all materials.\n\n"
                    )
                if "RelativePermeability" not in m:
                    Console.PrintMessage("m: {}\n".format(m))
                    raise general_writer.WriteError(
                        "The relative permeability must be specified for all materials.\n\n"
                    )
                self.write.material(name, "Name", m["Name"])
                conductivity = self.write.convert(m["ElectricalConductivity"], "T^3*I^2/(L^3*M)")
                conductivity = round(conductivity, 10)  # to get rid of numerical artifacts
                self.write.material(name, "Electric Conductivity", conductivity)
                self.write.material(
                    name, "Relative Permeability",
                    float(m["RelativePermeability"])
                )
                # permittivity might be necessary for the post processor
                if "RelativePermittivity" in m:
                    self.write.material(
                        name, "Relative Permittivity",
                        float(m["RelativePermittivity"])
                    )

    def _outputMagnetodynamic2DBodyForce(self, obj, name, equation):
        if hasattr(obj, "CurrentDensity_re_1"):
            # output only if current density is enabled and needed
            if not obj.CurrentDensity_re_1_Disabled:
                currentDensity = float(obj.CurrentDensity_re_1.getValueAs("A/m^2"))
                self.write.bodyForce(name, "Current Density", round(currentDensity, 6))
            # imaginaries are only needed for harmonic equation
            if equation.IsHarmonic:
                if not obj.CurrentDensity_im_1_Disabled:
                    currentDensity = float(obj.CurrentDensity_im_1.getValueAs("A/m^2"))
                    self.write.bodyForce(name, "Current Density Im", round(currentDensity, 6))

        if hasattr(obj, "Magnetization_re_1"):
            # output only if magnetization is enabled and needed
            if not obj.Magnetization_re_1_Disabled:
                magnetization = float(obj.Magnetization_re_1.getValueAs("A/m"))
                self.write.material(name, "Magnetization 1", round(magnetization, 6))
            if not obj.Magnetization_re_2_Disabled:
                magnetization = float(obj.Magnetization_re_2.getValueAs("A/m"))
                self.write.material(name, "Magnetization 2", round(magnetization, 6))
            # imaginaries are only needed for harmonic equation
            if equation.IsHarmonic:
                if not obj.Magnetization_im_1_Disabled:
                    magnetization = float(obj.Magnetization_im_1.getValueAs("A/m"))
                    self.write.material(name, "Magnetization Im 1", round(magnetization, 6))
                if not obj.Magnetization_im_2_Disabled:
                    magnetization = float(obj.Magnetization_im_2.getValueAs("A/m"))
                    self.write.material(name, "Magnetization Im 2", round(magnetization, 6))

    def handleMagnetodynamic2DBodyForces(self, bodies, equation):
        currentDensities = self.write.getMember("Fem::ConstraintCurrentDensity")
        for obj in currentDensities:
            if obj.References:
                for name in obj.References[0][1]:
                    self._outputMagnetodynamic2DBodyForce(obj, name, equation)
                self.write.handled(obj)
            else:
                # if there is only one current density without a reference,
                # add it to all bodies
                if len(currentDensities) == 1:
                    for name in bodies:
                        self._outputMagnetodynamic2DBodyForce(obj, name, equation)
                else:
                    raise general_writer.WriteError(
                        "Several current density constraints found without reference to a body.\n"
                        "Please set a body for each current density constraint."
                    )
            self.write.handled(obj)

        magnetizations = self.write.getMember("Fem::ConstraintMagnetization")
        for obj in magnetizations:
            if obj.References:
                for name in obj.References[0][1]:
                    self._outputMagnetodynamic2DBodyForce(obj, name, equation)
                self.write.handled(obj)
            else:
                # if there is only one magnetization without a reference,
                # add it to all bodies
                if len(magnetizations) == 1:
                    for name in bodies:
                        self._outputMagnetodynamic2DBodyForce(obj, name, equation)
                else:
                    raise general_writer.WriteError(
                        "Several magnetization constraints found without reference to a body.\n"
                        "Please set a body for each current density constraint."
                    )
            self.write.handled(obj)

    def handleMagnetodynamic2DBndConditions(self):
        for obj in self.write.getMember("Fem::ConstraintElectrostaticPotential"):
            if obj.References:
                for name in obj.References[0][1]:
                    # output the FreeCAD label as comment
                    if obj.Label:
                        self.write.boundary(name, "! FreeCAD Name", obj.Label)
                    if obj.PotentialEnabled:
                        if hasattr(obj, "Potential"):
                            potential = float(obj.Potential.getValueAs("V"))
                            self.write.boundary(name, "Potential", round(potential, 6))
                    if obj.ElectricInfinity:
                        self.write.boundary(name, "Infinity BC", True)
                self.write.handled(obj)

    def handleMagnetodynamic2DEquation(self, bodies, equation):
        for b in bodies:
            if equation.IsHarmonic and (equation.AngularFrequency == 0):
                raise general_writer.WriteError(
                    "The angular frequency must not be zero.\n\n"
                )
            self.write.equation(b, "Name", equation.Name)
            if equation.IsHarmonic:
                frequency = float(Units.Quantity(equation.AngularFrequency).Value)
                self.write.equation(b, "Angular Frequency", round(frequency, 6))

##  @}



# ***************************************************************************
# *   Copyright (c) 2014 sliptonic <shopinthewoods@gmail.com>               *
# *   Copyright (c) 2022 Larry Woestman <LarryWoestman2@gmail.com>          *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Lesser General Public License for more details.                   *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************


import Path.Post.UtilsArguments as PostUtilsArguments
import Path.Post.UtilsExport as PostUtilsExport

#
# The following variables need to be global variables
# to keep the PathPostProcessor.load method happy:
#
#    TOOLTIP
#    TOOLTIP_ARGS
#    UNITS
#
#    The "argument_defaults", "arguments_visible", and the "values" hashes
#    need to be defined before the "init_shared_arguments" routine can be
#    called to create TOOLTIP_ARGS, so they also end up having to be globals.
#
TOOLTIP = """This is a postprocessor file for the Path workbench. It is used to
take a pseudo-G-code fragment outputted by a Path object, and output
real G-code suitable for a linuxcnc 3 axis mill. This postprocessor, once placed
in the appropriate PathScripts folder, can be used directly from inside
FreeCAD, via the GUI importer or via python scripts with:

import refactored_linuxcnc_post
refactored_linuxcnc_post.export(object,"/path/to/file.ncc","")
"""
#
# Default to metric mode
#
UNITS = "G21"


def init_values(values):
    """Initialize values that are used throughout the postprocessor."""
    #
    global UNITS

    PostUtilsArguments.init_shared_values(values)
    #
    # Set any values here that need to override the default values set
    # in the init_shared_values routine.
    #
    values["ENABLE_COOLANT"] = True
    # the order of parameters
    # linuxcnc doesn't want K properties on XY plane; Arcs need work.
    values["PARAMETER_ORDER"] = [
        "X",
        "Y",
        "Z",
        "A",
        "B",
        "C",
        "I",
        "J",
        "F",
        "S",
        "T",
        "Q",
        "R",
        "L",
        "H",
        "D",
        "P",
    ]
    #
    # Used in the argparser code as the "name" of the postprocessor program.
    # This would normally show up in the usage message in the TOOLTIP_ARGS,
    # but we are suppressing the usage message, so it doesn't show up after all.
    #
    values["MACHINE_NAME"] = "LinuxCNC"
    #
    # Any commands in this value will be output as the last commands
    # in the G-code file.
    #
    values[
        "POSTAMBLE"
    ] = """M05
G17 G54 G90 G80 G40
M2"""
    values["POSTPROCESSOR_FILE_NAME"] = __name__
    #
    # Any commands in this value will be output after the header and
    # safety block at the beginning of the G-code file.
    #
    values["PREAMBLE"] = """G17 G54 G40 G49 G80 G90"""
    values["UNITS"] = UNITS


def init_argument_defaults(argument_defaults):
    """Initialize which arguments (in a pair) are shown as the default argument."""
    PostUtilsArguments.init_argument_defaults(argument_defaults)
    #
    # Modify which argument to show as the default in flag-type arguments here.
    # If the value is True, the first argument will be shown as the default.
    # If the value is False, the second argument will be shown as the default.
    #
    # For example, if you want to show Metric mode as the default, use:
    #   argument_defaults["metric_inch"] = True
    #
    # If you want to show that "Don't pop up editor for writing output" is
    # the default, use:
    #   argument_defaults["show-editor"] = False.
    #
    # Note:  You also need to modify the corresponding entries in the "values" hash
    #        to actually make the default value(s) change to match.
    #


def init_arguments_visible(arguments_visible):
    """Initialize which argument pairs are visible in TOOLTIP_ARGS."""
    PostUtilsArguments.init_arguments_visible(arguments_visible)
    #
    # Modify the visibility of any arguments from the defaults here.
    #


def init_arguments(values, argument_defaults, arguments_visible):
    """Initialize the shared argument definitions."""
    parser = PostUtilsArguments.init_shared_arguments(
        values, argument_defaults, arguments_visible
    )
    #
    # Add any argument definitions that are not shared with all other
    # postprocessors here.
    #
    return parser


#
# Creating global variables and using functions to modify them
# is useful for being able to test things later.
#
values = {}
init_values(values)
argument_defaults = {}
init_argument_defaults(argument_defaults)
arguments_visible = {}
init_arguments_visible(arguments_visible)
parser = init_arguments(values, argument_defaults, arguments_visible)
#
# The TOOLTIP_ARGS value is created from the help information about the arguments.
#
TOOLTIP_ARGS = parser.format_help()
#
# Create another parser just to get a list of all possible arguments
# that may be output using --output_all_arguments.
#
all_arguments_visible = {}
for k in iter(arguments_visible):
    all_arguments_visible[k] = True
all_visible = init_arguments(values, argument_defaults, all_arguments_visible)


def export(objectslist, filename, argstring):
    """Postprocess the objects in objectslist to filename."""
    #
    global parser
    global UNITS
    global values

    # print(parser.format_help())

    (flag, args) = PostUtilsArguments.process_shared_arguments(
        values, parser, argstring, all_visible, filename
    )
    if not flag:
        return args
    #
    # Process any additional arguments here
    #

    #
    # Update the global variables that might have been modified
    # while processing the arguments.
    #
    UNITS = values["UNITS"]

    return PostUtilsExport.export_common(values, objectslist, filename)



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2022-2023 FreeCAD Project Association                   *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

import addonmanager_freecad_interface as fci

# Unit tests for the Addon Manager module
from AddonManagerTest.app.test_utilities import (
    TestUtilities as AddonManagerTestUtilities,
)
from AddonManagerTest.app.test_addon import (
    TestAddon as AddonManagerTestAddon,
)
from AddonManagerTest.app.test_macro import (
    TestMacro as AddonManagerTestMacro,
)
from AddonManagerTest.app.test_git import (
    TestGit as AddonManagerTestGit,
)
from AddonManagerTest.app.test_installer import (
    TestAddonInstaller as AddonManagerTestAddonInstaller,
    TestMacroInstaller as AddonManagerTestMacroInstaller,
)
from AddonManagerTest.app.test_dependency_installer import (
    TestDependencyInstaller as AddonManagerTestDependencyInstaller,
)
from AddonManagerTest.app.test_uninstaller import (
    TestAddonUninstaller as AddonManagerTestAddonUninstaller,
    TestMacroUninstaller as AddonManagerTestMacroUninstaller,
)
from AddonManagerTest.app.test_freecad_interface import (
    TestConsole as AddonManagerTestConsole,
    TestParameters as AddonManagerTestParameters,
    TestDataPaths as AddonManagerTestDataPaths,
)
from AddonManagerTest.app.test_metadata import (
    TestDependencyType as AddonManagerTestDependencyType,
    TestMetadataReader as AddonManagerTestMetadataReader,
    TestMetadataReaderIntegration as AddonManagerTestMetadataReaderIntegration,
    TestUrlType as AddonManagerTestUrlType,
    TestVersion as AddonManagerTestVersion,
    TestMetadataAuxiliaryFunctions as AddonManagerTestMetadataAuxiliaryFunctions,
)


class TestListTerminator:
    pass


# Basic usage mostly to get static analyzers to stop complaining about unused imports
try:
    import FreeCAD
except ImportError:
    FreeCAD = None
loaded_gui_tests = [
    AddonManagerTestUtilities,
    AddonManagerTestAddon,
    AddonManagerTestMacro,
    AddonManagerTestGit,
    AddonManagerTestAddonInstaller,
    AddonManagerTestMacroInstaller,
    AddonManagerTestDependencyInstaller,
    AddonManagerTestAddonUninstaller,
    AddonManagerTestMacroUninstaller,
    AddonManagerTestConsole,
    AddonManagerTestParameters,
    AddonManagerTestDataPaths,
    AddonManagerTestDependencyType,
    AddonManagerTestMetadataReader,
    AddonManagerTestMetadataReaderIntegration,
    AddonManagerTestUrlType,
    AddonManagerTestVersion,
    AddonManagerTestMetadataAuxiliaryFunctions,
    TestListTerminator,  # Needed to prevent the last test from running twice
]
for test in loaded_gui_tests:
    fci.Console.PrintLog(f"Loaded tests from {test.__name__}\n")



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2022-2023 FreeCAD Project Association                   *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

import unittest
import os
import FreeCAD
from addonmanager_workers_utility import ConnectionChecker
from PySide import QtCore

import NetworkManager


class TestWorkersUtility(unittest.TestCase):

    MODULE = "test_workers_utility"  # file name without extension

    @unittest.skip("Test is slow and uses the network: refactor!")
    def setUp(self):
        self.test_dir = os.path.join(
            FreeCAD.getHomePath(), "Mod", "AddonManager", "AddonManagerTest", "data"
        )
        self.last_result = None

        url = "https://api.github.com/zen"
        NetworkManager.InitializeNetworkManager()
        result = NetworkManager.AM_NETWORK_MANAGER.blocking_get(url)
        if result is None:
            self.skipTest("No active internet connection detected")

    def test_connection_checker_basic(self):
        """Tests the connection checking worker's basic operation: does not exit until worker thread completes"""
        worker = ConnectionChecker()
        worker.success.connect(self.connection_succeeded)
        worker.failure.connect(self.connection_failed)
        self.last_result = None
        worker.start()
        while worker.isRunning():
            QtCore.QCoreApplication.processEvents(QtCore.QEventLoop.AllEvents, 50)
        QtCore.QCoreApplication.processEvents(QtCore.QEventLoop.AllEvents)
        self.assertEqual(self.last_result, "SUCCESS")

    def test_connection_checker_thread_interrupt(self):
        worker = ConnectionChecker()
        worker.success.connect(self.connection_succeeded)
        worker.failure.connect(self.connection_failed)
        self.last_result = None
        worker.start()
        worker.requestInterruption()
        while worker.isRunning():
            QtCore.QCoreApplication.processEvents(QtCore.QEventLoop.AllEvents, 50)
        QtCore.QCoreApplication.processEvents(QtCore.QEventLoop.AllEvents)
        self.assertIsNone(
            self.last_result, "Requesting interruption of thread failed to interrupt"
        )

    def connection_succeeded(self):
        self.last_result = "SUCCESS"

    def connection_failed(self):
        self.last_result = "FAILURE"



#***************************************************************************
#*   Copyright (c) 2011 Juergen Riegel <FreeCAD@juergen-riegel.net>        *
#*                                                                         *
#*   This program is free software; you can redistribute it and/or modify  *
#*   it under the terms of the GNU Lesser General Public License (LGPL)    *
#*   as published by the Free Software Foundation; either version 2 of     *
#*   the License, or (at your option) any later version.                   *
#*   for detail see the LICENCE text file.                                 *
#*                                                                         *
#*   This program is distributed in the hope that it will be useful,       *
#*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
#*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
#*   GNU Library General Public License for more details.                  *
#*                                                                         *
#*   You should have received a copy of the GNU Library General Public     *
#*   License along with this program; if not, write to the Free Software   *
#*   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
#*   USA                                                                   *
#*                                                                         *
#***************************************************************************

from __future__ import division
from math import pi
import unittest

import FreeCAD

class TestFillet(unittest.TestCase):
    def setUp(self):
        self.Doc = FreeCAD.newDocument("PartDesignTestFillet")

    def testFilletCubeToSphere(self):
        self.Body = self.Doc.addObject('PartDesign::Body','Body')
        self.Box = self.Doc.addObject('PartDesign::AdditiveBox','Box')
        self.Body.addObject(self.Box)
        self.Box.Length=10.00
        self.Box.Width=10.00
        self.Box.Height=10.00
        self.Doc.recompute()
        self.Fillet = self.Doc.addObject("PartDesign::Fillet","Fillet")
        self.Fillet.Base = (self.Box, ['Face'+str(i+1) for i in range(6)])
        self.Fillet.Radius = 4.999999
        self.Body.addObject(self.Fillet)
        self.Doc.recompute()
        self.assertAlmostEqual(self.Fillet.Shape.Volume, 4/3 * pi * 5**3, places=3)
        #test UseAllEdges property
        self.Fillet.UseAllEdges = True
        self.Fillet.Base = (self.Box, ['']) # no subobjects, should still work
        self.Doc.recompute()
        self.assertAlmostEqual(self.Fillet.Shape.Volume, 4/3 * pi * 5**3, places=3)
        self.Fillet.Base = (self.Box, ['Face50']) # non-existent face, topo naming resilience
        self.Doc.recompute()
        self.assertAlmostEqual(self.Fillet.Shape.Volume, 4/3 * pi * 5**3, places=3)
        self.Fillet.UseAllEdges = False
        self.Fillet.Base = (self.Box, ['Face1'])
        self.Doc.recompute()
        self.assertNotAlmostEqual(self.Fillet.Shape.Volume, 4/3 * pi * 5**3, places=3)

    def tearDown(self):
        #closing doc
        FreeCAD.closeDocument("PartDesignTestFillet")
        # print ("omit closing document for debugging")

